/**
 * Camera.java
 *
 * This acts as a  "simple" layer to interface with the camera and it's functions. It'll run (or start) all the appropriate
 * image processing and suchlike, and is mainly an interface/container class, no actual planning or anything is done here (like
 * finding the car etc.).
 * 
 * The actual camera image might be created as a GUI element before this exists, so be careful when to mess with this.
 * 
 * This does however handle if the camera is on, and handles the thread ticks, which were moved out of the engine.
 * 
 * Created by; Andrew Armstrong
 * Last Modified: 04/05/08
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.lang.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.*;

public class Camera implements Runnable {

	// Is the camera on? yes/no
	boolean cameraOn;
	// GUI work needs this
	private CarAIMain parentCarAIMain;

	// stored reference image, used for manual controls for getting frame differencing etc.
	BufferedImage referenceImage;
	
	// Image processing stuff
	ImageThresholder imageThresholder;
	BackgroundSubtracter backgroundSubtracter;
	
	// This makes sure the thread knows if we have attempted to start the camera yet
	int cameraAttemptCounter = 0;

	// Thread for the camera
	Thread camThread;
	
	public Camera(CarAIMain passedCarAIMain) {
		parentCarAIMain = passedCarAIMain;
		
		// Camera starts off...
		cameraOn = false;

		// Create the image processing classes
		imageThresholder = new ImageThresholder(passedCarAIMain);
		backgroundSubtracter = new BackgroundSubtracter(passedCarAIMain);
	}

	// Threshold button action
	public void thresholdButton() {
		// Get fresh threshold image from live image
		BufferedImage thresholdImage = imageThresholder.applyThresholdToImage();

		// Put in debug location
		parentCarAIMain.staticCamImg.image.setImage(thresholdImage);
	}

	// Background Subtraction button
	public void backgroundSubtractionButton() {	
		// Get Live image
		BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();

		// Apply BG Sub
		BufferedImage bgSubImage = backgroundSubtracter.applyBackgroundSubtraction(liveImage, referenceImage);

		// Put in debug location
		parentCarAIMain.staticCamImg.image.setImage(bgSubImage);
	}
	
	// Saves the current debug image as the reference
	public void saveReferenceButton() {
		// Save the static image as a new copy
		referenceImage = copyImage(parentCarAIMain.staticCamImg.image.getImage());
	}

	// Frame differencing button
	public void frameDifferencing() {
		// Do thresholding first on new live image
		BufferedImage thresholdImage = imageThresholder.applyThresholdToImage();
		BufferedImage frameDifferenced = FrameDifferencing.frameDifference(referenceImage, thresholdImage);
		
		// Put in debug location
		parentCarAIMain.staticCamImg.image.setImage(frameDifferenced);
	}
	
	// Test function for reporting a 5x5 square of image, starting from (0,9) pixels on the image, to get the colour of something
	// This will only ever be run from a test button, left in case debugging is needed
	void Test_CheckTopLeftColour()
	{
		BufferedImage image = parentCarAIMain.camImg.getLiveImage();
	
		// Vars
		int rgb=0, r=0, g=0, b=0;
			
		// Loop
		for (int x = 0; x < 5; x++)
		{
			for (int y = 9; y < 14; y++)
			{
				// Within a threshold of colour
				rgb = image.getRGB(x, y);
				
				// Red
				r = (rgb & 0xFF0000) >> 16;
				g = (rgb & 0x00FF00) >> 8;
				b = (rgb & 0x0000FF);
				System.out.println("[Camera] Test_CheckTopLeftColour (" + x + "," + y + ") getRGB: [" + rgb + "] R,G,B: (" + r + "," + g + "," + b + ")");
			
				image.setRGB(x, y, Color.RED.getRGB());
			}
		}
		
		// Debug
		parentCarAIMain.staticCamImg.image.setImage(image);
	}

	// Draws a block marker on a buffered image at x, y
	// Default colour is set to red
	public void drawBlock(Coordinate coords, BufferedImage image) { drawBlock(coords.x, coords.y, image, Color.RED); }
	public void drawBlock(Coordinate coords, BufferedImage image, Color color) { drawBlock(coords.x, coords.y, image, color); }
	public void drawBlock(int x, int y, BufferedImage image) { drawBlock(x, y, image, Color.RED); }
	public void drawBlock(int x, int y, BufferedImage image, Color color)
	{
		int i, j; // i = j, j == x
		int imageHeight = image.getHeight();
		int imageWidth = image.getWidth();
		if(x >= 0 && y >= 0 && x < imageWidth && y < imageHeight)
		{
			for(i = x - 3; i < x + 3; i++)
			{
				i = Utility.getBoundriedInt(i, 0, imageWidth);
				for(j = y - 3; j < y+3; j++)
				{
					j = Utility.getBoundriedInt(j, 0, imageHeight);
					image.setRGB(i, j, color.getRGB());
				}
			}
		}
		else
		{
			parentCarAIMain.log("[Camera] Couldn't draw block, invalid X and Y co-ordinates.");
		}
	}

	// Draws a horizontal line on the Y co-ordinate given
	// Default colour is set to red
	public void drawHorizontalLine(int y, BufferedImage image) { drawHorizontalLine(y, image, Color.RED); }
	public void drawHorizontalLine(int y, BufferedImage image, Color color)
	{
		int imageWidth = image.getWidth();
		if(y >= 0 && y < image.getHeight())
		{
			for(int x = 0; x < imageWidth; x++)
			{
				image.setRGB(x, y, color.getRGB());
			}
		}
		else
		{
			parentCarAIMain.log("[Camera] Couldn't draw line, invalid Y co-ordinate.");
		}
	}
	// Draws a vertical line on the X co-ordinate given
	// Default colour is set to red
	public void drawVerticalLine(int x, BufferedImage image) { drawVerticalLine(x, image, Color.RED); }
	public void drawVerticalLine(int x, BufferedImage image, Color color)
	{
		int imageHeight = image.getHeight();
		if(x >= 0 && x < image.getWidth())
		{
			for(int y = 0; y < imageHeight; y++)
			{
				image.setRGB(x, y, color.getRGB());
			}
		}
		else
		{
			parentCarAIMain.log("[Camera] Couldn't draw line, invalid Y co-ordinate.");
		}
	}
	// Draws a retangle starting in the top left (x, y) and going for x width and y height. Only do outline!
	// Default colour is set to blue
	public void drawRectangle(int x, int y, int width, int height, BufferedImage image) { drawRectangle(x, y, width, height, image, Color.BLUE); }
	public void drawRectangle(int x, int y, int width, int height, BufferedImage image, Color color)
	{
		int colorInt =  color.getRGB();
		int xSub = 0, ySub = 0;
		
		// Top	
		for(xSub = x; xSub <= x + width; xSub++) {
			image.setRGB(xSub, y, colorInt);				
		}
		// Bottom
		for(xSub = x; xSub <= x + width; xSub++) {
			image.setRGB(xSub, y + height, colorInt);				
		}
		// Left
		for(ySub = y; ySub <= y + height; ySub++) {
			image.setRGB(x, ySub, colorInt);				
		}
		// Right
		for(ySub = y; ySub <= y + height; ySub++) {
			image.setRGB(x + width, ySub, colorInt);				
		}
	}
		
	// Draw a line
	public void drawLine(Coordinate i, Coordinate j, BufferedImage image, Color color)
	{
		drawLine(i.x, i.y, j.x, j.y, image, color);
	}
	public void drawLine(int xi, int yi, int xj, int yj, BufferedImage image, Color color)
	{
		// Use a graphics 2d object to draw it using java in-built functions
		Graphics2D g = image.createGraphics();
		g.setColor(color);
		g.drawLine(xi, yi, xj, yj);
	}

	// Copy image if needed. The way Java works is that any object (which for this class mainly means "BufferedImage" classes) is passed
	// as a reference unless specifically created as a "new" instance.
	public BufferedImage copyImage(BufferedImage image)
	{
		BufferedImage newImage = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
		for (int y=0; y<image.getHeight(); y++)	// scan columns
			for (int x=0; x<image.getWidth(); x++) // scan pixels in current row
				newImage.setRGB(x,y,image.getRGB(x,y));	// copy pixel
		return newImage;
	}
	
	// Report whether camera is enabled
	public boolean getCameraOn(){
		return cameraOn;
	}

	// Enable/disable camera
	public void enableCam(){
		// Begin thread to initiate camera (thread allows status messages to appear at correct interval)
		camThread = new Thread(this);
		camThread.start();
	}
	public void disableCam(){
		// Disables the camera, turns it off
		cameraAttemptCounter = 0;
		parentCarAIMain.camImg.disableCam();
		parentCarAIMain.log("[Camera] Camera Disabled");
		cameraOn = false;
		parentCarAIMain.camButton.setText("Turn On Camera");
	}
	
	// Load and Save camera images
	public void saveLiveCameraImage() {
		if(getCameraOn())
		{	
			// Save the current image (if we have one) to a variable
			BufferedImage save_image = parentCarAIMain.camImg.liveImage.getImage();
			
			// Ask for where it should be stored, with what name
			final JFileChooser fc = new JFileChooser();
			// In response to a button click
			int returnVal = fc.showSaveDialog(parentCarAIMain.saveLiveImageButton);//, "Choose a image file to open");

			if (returnVal == JFileChooser.APPROVE_OPTION) {
				File file = fc.getSelectedFile();
				parentCarAIMain.log("[Camera] Saving live image: " + file.getName() + ".");
				
				// Output the image to that location as a PNG
				// Write generated image to a file
				try {
					// Save as PNG
					//File file = new File("newimage.png");
					ImageIO.write(save_image, "png", file);
				}
				catch (IOException e) {
					parentCarAIMain.log("[Camera] Error writing PNG file. Error: " + e.toString());
				}
			} else {
				parentCarAIMain.log("[Camera] Save command cancelled by user.");
			}
		}
		else{
			parentCarAIMain.log("[Camera] Camera off: Cannot save image until enabled.");
		}
	}
	
	// Save static
	public void saveStaticCameraImage() {
		// Save the current image (if we have one) to a variable
		BufferedImage save_image = parentCarAIMain.staticCamImg.image.getImage();

		// Ask for where it should be stored, with what name
		final JFileChooser fc = new JFileChooser();
		// In response to a button click
		int returnVal = fc.showSaveDialog(parentCarAIMain.saveStaticImageButton);//, "Choose a image file to open");

		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			parentCarAIMain.log("[Camera] Saving static image: " + file.getName() + ".");
			
			// Output the image to that location as a PNG
			// Write generated image to a file
			try {
				// Save as PNG
				//File file = new File("newimage.png");
				ImageIO.write(save_image, "png", file);
			}
			catch (IOException e) {
				parentCarAIMain.log("[Camera] Error writing PNG file. Error: " + e.toString());
			}
		} else {
			parentCarAIMain.log("[Camera] Save command cancelled by user.");
		}
	}
	
	// - Maybe only alter the static image?
	public void loadCameraImage() {
		// Popup a "Load picture" box
		// Ask for where it should be stored, with what name
		final JFileChooser fc = new JFileChooser();
		// In response to a button click
		// parentCarAIMain.loadImageButton.getParent()
		int returnVal = fc.showOpenDialog(parentCarAIMain.loadImageButton);//, "Choose a image file to open");
		//fc.setFileView(new ImageFileView());

		if (returnVal == JFileChooser.APPROVE_OPTION) {
			File file = fc.getSelectedFile();
			//This is where a real application would open the file.
			parentCarAIMain.log("[Camera] Opening: " + file.getName() + ".");
			
			// Attempt to load this as a png image
			BufferedImage load_image = null;
		    try {
		        // Read from a PNG file
		        load_image = ImageIO.read(file);
		    }
			catch (IOException e) {
				parentCarAIMain.log("[Camera] Error reading file: [" + file.getName() + "] Error: " + e.toString());
		    }
		    
			// Put it in the static camera bufferimage
			parentCarAIMain.staticCamImg.image.setImage(load_image);
			// Old, set as live image
			//parentCarAIMain.camImg.liveImage.setImage(load_image);
		} else {
			parentCarAIMain.log("[Camera] Open command cancelled by user.");
		}
	}
	
	// Thread for turning on the camera
	public void run()
	{
		// Determine which Thread is currently running
		Thread thisThread = Thread.currentThread();

		//Thread actions when initialising camera
		while(thisThread == camThread){

			String camName = parentCarAIMain.camImg.CAPTURE_DEVICE_NAME;

			// Thread actions when initialising camera
			while(thisThread == camThread){
				if(cameraAttemptCounter == 0){
					parentCarAIMain.camButton.setEnabled(false);
					parentCarAIMain.log("[Camera] Initialising video capture device: "+ camName + "\nPlease Wait...");
					cameraAttemptCounter++;
					try{
						Thread.sleep(1000);
					}
					catch(InterruptedException e) {
						// This should cirtainly never happen, surely?
						parentCarAIMain.log("[Camera] Error sleeping camera thread. Error: " + e.toString());
					}
				}
				else{
					try{
						parentCarAIMain.camImg.enableCam(parentCarAIMain);
						parentCarAIMain.log("[Camera] Camera Enabled.");
						cameraOn = true;
						parentCarAIMain.camButton.setText("Turn Off Camera");
						parentCarAIMain.camButton.setEnabled(true);
					} catch(Exception e){
						parentCarAIMain.log("[Camera] Camera not found. Please connect camera correctly and restart app if you wish to use the camera.");
						parentCarAIMain.log("[Camera] Error occured: " + e.toString());
						//Not possible to have second attempt at connecting to camera (generates error/JRE bug), so disable button
						parentCarAIMain.camButton.setText("Camera Unavailable");
					}
					camThread = null;
				}
			}
		}
	}	
}