/**
 * CameraController.java
 *
 * Initialises and controls input from the camera. Called by CameraImage.java
 *
 * Original Code: David Bull - http://www.uk-dave.com/projects/java/colourtracker.php
 * Modified by: Scott Culcheth
 * Further Modified by: Chris Carter
 *
 * Last Modified: 24.04.07
 *
 */

import java.awt.*;
import java.awt.image.*;
import javax.media.*;
import javax.media.control.*;
import javax.media.format.*;
import javax.media.protocol.*;
import javax.media.util.*;

public class CameraController
{
	////////////////////////////////////////////////////
	// CONSTANTS AND CLASS VARIABLES
	////////////////////////////////////////////////////

	private Player player;
	private String msg;
	private CarAIMain parentCarAIMain;

	////////////////////////////////////////////////////
	// CONSTRUCTOR
	////////////////////////////////////////////////////

	/**
	 * Initialises the video capture device.
	 * @param name the name of the device (e.g. <code>vfw:Microsoft WDM Image Capture (Win32):0</code>).
	 * @param locator the location of the device (e.g. <code>vfw://0</code>).
	 * @param size
	 * @throws Exception if there was a problem accessing the camera.
	 */
	public CameraController(String name, String locator, Dimension size, CarAIMain passedCarAIMain) //throws Exception
	{
		// Create capture device
		parentCarAIMain = passedCarAIMain;
		try
		{
			CaptureDeviceInfo captureDeviceInfo = CaptureDeviceManager.getDevice(name);
			VideoFormat videoFormat = null;
			Format[] formats = captureDeviceInfo.getFormats();
			for (int i=0; i<formats.length; i++)
			{
				if (((VideoFormat)formats[i]).getSize().equals(size))
				{
					videoFormat = (VideoFormat)formats[i];
					break;
				}
			}
			DataSource dataSource = Manager.createDataSource(new MediaLocator(locator));
			FormatControl[] formatControls = ((CaptureDevice)dataSource).getFormatControls();
			for (int i=0; i<formatControls.length; i++)
			{
				if (formatControls[i]!=null)
				{
					Format supportedFormats[] = formatControls[i].getSupportedFormats();
					if (supportedFormats!=null)
						for (int j=0;  j<supportedFormats.length;  j++)
							if (supportedFormats[j].matches(videoFormat))
								formatControls[i].setFormat(videoFormat);
				}
			}
			CaptureDeviceInfo deviceInfo = ((CaptureDevice)dataSource).getCaptureDeviceInfo();
			((CaptureDevice)dataSource).disconnect();
			player = Manager.createRealizedPlayer(deviceInfo.getLocator());
			player.start();

			try { Thread.sleep(3000); }
			catch (InterruptedException e) {
				parentCarAIMain.log("Error sleeping thread for camera starting. Error: " + e.toString());
			}
		}
		catch (Exception e)
		{
			parentCarAIMain.log("Error Initialising video capture device! Error: " + e.toString());
			//throw e;
		}
	}



	////////////////////////////////////////////////////
	// METHODS
	////////////////////////////////////////////////////

	/**
	 * Shutsdown the video capture device.
	 * Calls to other methods after calling shutdown will not succeed.
	 * Note that this method MUST be called when exiting the application.
	 */
	public void shutdown()
	{
		player.close();
		player.deallocate();
	}


	/**
	 * Grabs a frame from the video device.
	 * @return a <code>BufferedImage</code> containing the grabbed frame.
	 */
	public BufferedImage getFrame()
	{
		// Grab a frame from the capture device
		FrameGrabbingControl frameGrabber = (FrameGrabbingControl)player.getControl("javax.media.control.FrameGrabbingControl");
		Buffer buf = frameGrabber.grabFrame();
		// Convert frame to an buffered image and return
		Image img = (new BufferToImage((VideoFormat)buf.getFormat()).createImage(buf));
		if (img==null)
		{
			parentCarAIMain.log("Error: Capture device doesnt appear to be initialised yet");
			return null;		// happens if video device isn't properly initialised yet
		}
		BufferedImage buffImg = new BufferedImage(img.getWidth(null), img.getHeight(null), BufferedImage.TYPE_INT_RGB);
		Graphics2D g = buffImg.createGraphics();
		g.drawImage(img, null, null);
		return buffImg;

	}


	/**
	 * Returns the visual component, allowing the video stream from the capture device to be shown onscreen.
	 * @return the visual component.
	 */
	public Component getVisualComponent()
	{
		return player.getVisualComponent();
	}

	public float getMaxFrameRate()
	{
		FrameRateControl frameRate = (FrameRateControl)player.getControl("javax.media.control.FrameRateControl");
		return frameRate.getMaxSupportedFrameRate();
	}
}
