/**
 * Car.java
 * 
 * Finds the car, tracks where it was previously with variables, interfaced
 * by NavigationAI.java
 *
 * This works closely with camera objects passed to it.
 *
 * Created By: Andrew Armstrong
 * Last Modified: 04/05/08
 */
import java.util.*;
import java.awt.*;
import java.awt.image.*;

 
public class Car {

	/*************************
		Variables
	*************************/

	// Variables for what we alter the final image to be - empty is black, full is green.
	public static final int EMPTY_REGION = 0x000000;
	public static final int FULL_REGION  = 0x00FF00;

	public static final int MATCH_EMPTY_REGION = EMPTY_REGION - 16777216; // For getRGB, magic number removed!
	public static final int MATCH_FULL_REGION  = FULL_REGION - 16777216; // For getRGB, magic number removed!

	// Track limits are used often to cut down CPU usage. Set via setTrackSize.
	private int trackMinX = 0, trackMinY = 0, trackMaxX = 320, trackMaxY = 240;

	// Memory
	double lastCarFacing; // Degree of car facing from our last frame. Facing "Right" on the image is 0 (down is 90...etc)
	Coordinate lastCarCoords; // Centre co-ordinates of the car's last  known position

	// For logging and ease of use
	private CarAIMain parentCarAIMain;
	private Camera camera;
	
	/*************************
		Constructor
	*************************/

	public Car(CarAIMain passedCarAIMain, Camera passedCamera)
	{
		passedCarAIMain.logDetail("[Car] Constructor: Starting");
		parentCarAIMain = passedCarAIMain;
		this.camera = passedCamera;
		lastCarCoords = new Coordinate();
		passedCarAIMain.logDetail("[Car] Constructor: Done");
	}

	/*************************
		PUBLIC FUNCTIONS 
	*************************/
	
	// setTrackSize
	// This simplifies me splitting/merging some functions. Track.java sorts the track, and gets passed this Car class to call this, so we
	// can set the track variables stored above
	public void setTrackSize(int minX, int maxX, int minY, int maxY) {
		trackMinX = minX;
		trackMaxX = maxX;
		trackMinY = minY;
		trackMaxY = maxY;
	}
	
	// getCarPosition
	// Will get the position of the car wherever it is on the track, returns true if successful, else false
	// This doesn't need the last car coordinates.
	public boolean getCarPosition(BufferedImage liveImage, BufferedImage image) {
		// Require good images
		if(liveImage == null || image == null || liveImage.getWidth() == 0 || image.getWidth() == 0)
		{
			parentCarAIMain.log("[Car] getCarPosition - Error, no or invalid reference image.");
			return false;
		}		
		if(liveImage.getWidth() != image.getWidth())
		{
			parentCarAIMain.log("[Car] getCarPosition - Error, images are of difference sizes.");
			return false;
		}
		// Now search the image. We for now work on a x=0, y=0 to the max loop.
		int xCar = 0, yCar = 0, matches = 0;
		// This is the area size (2 * countPixelAreaSize * 2 * countPixelAreaSize total pixels) around a full pixel we look to see if there is enough to consider the car being part of the area
		int countPixelAreaSize = 3;
		// This is the area we use to average and find the centre of the car location. We find quickly a tiny location which should be it, and average all nearby points of relevant "fullness" to get the general centre
		// Experiment with this number if it isn't entirely accurate all the time.
		int averageLocationAreaSize = 17;		
		// This should be set to the "minimum" area that can be considered part of the car. For performance, set lower.
		int minimumValidPoints = 20;
		int highestmatches = 0;

		parentCarAIMain.logDetail("[Car] getCarPosition. Looking for car. Max X and Y: (" + trackMaxX + "," + trackMaxY + "). Full: " + FULL_REGION + ". Empty: " + EMPTY_REGION);

		// Loop the pixels
		for (int x = trackMinX; x < trackMaxX; x++)
		{
			for (int y = trackMinY; y < trackMaxY; y++)
			{
				// Get the RGB values here. If full, we will search around
				if(image.getRGB(x,y) == MATCH_FULL_REGION)
				{
					// Check the pixels in this area (+/- 6 from coordinates)
					matches = countPixelsInArea(x, y, countPixelAreaSize, MATCH_FULL_REGION, image);

					// Has it found a large enough area?
					if(matches > highestmatches)
					{
						// This is a place we want to investigate further!
						highestmatches = matches;
						xCar = x;
						yCar = y;
						
						// NB: Debug slows down the loop (true GUI logging really slows it down) so don't do it here unless you need to know what is found.
						//System.out.println("Green pixels around: (" + x + "," + y + ") total: " + matches);
					}
				}
			}
		}
		// Do we have something which has at least the minimum number of valid points?
		if(highestmatches > minimumValidPoints)
		{
			// Average out 
			Coordinate xyAverage = findAverageColorLocationInArea(xCar, yCar, averageLocationAreaSize, MATCH_FULL_REGION, image);

			// Now we have this car's location find it's facing
			double facing = findCarFacing(xyAverage.x, xyAverage.y, liveImage, image);
			
			// Errors?
			if(facing < 0)
			{
				// Cannot be negative, we have a problem. We simply don't know the facing!
				parentCarAIMain.logDetail("[Car] getCarPosition. Error finding facing we cannot get out of, please retry! Error occured.");
				return false;
			}
			// Set since no errors, we have the last location of the car
			lastCarCoords.x = xyAverage.x;
			lastCarCoords.y = xyAverage.y;
			
			// Got a centre location for our car, hopefully!
			parentCarAIMain.logDetail("[Car] getCarPosition. Found highest matching point with [" + highestmatches + "] matches at (" + xCar + "," + yCar + "). Average, and thus car centre is at (" + xyAverage.x + "," + xyAverage.y + "). Angle of car: " + facing);
			// Log Coordinates
			parentCarAIMain.logCoords("[Car] Car is at: (" + xCar + "," + yCar + ") facing [" + facing + "]");

			// Draw the car
			camera.drawBlock(xyAverage.x, xyAverage.y, image);
			// Put in debug location
			parentCarAIMain.staticCamImg.image.setImage(image);
			
			return true;
		}

		parentCarAIMain.log("[Car] getCarPosition. Not found any points surrounded by [" + minimumValidPoints + "] or more pixels of colour. Maximum is: " + highestmatches);
		// Put in debug location
		parentCarAIMain.staticCamImg.image.setImage(image);

		return false;
	}

	// Find the car around the past coordiates we stored
	public boolean findCarAroundPastCoordinates(BufferedImage liveImage, BufferedImage image)
	{
		// Add in x, y values
		return findCarAroundExistingPoint(liveImage, image, lastCarCoords.x, lastCarCoords.y, trackMinX, trackMaxX, trackMinY, trackMaxY);
	}
	
	// Finds a car object near x, y, with the minimum and maximums being the extreme of the track
	public boolean findCarAroundExistingPoint(BufferedImage liveImage, BufferedImage image, int x, int y)
	{
		// Automatic min/max X/Y values
		return findCarAroundExistingPoint(liveImage, image, x, y, trackMinX, trackMaxX, trackMinY, trackMaxY);
	}
	// Custom version could use smaller areas if needed
	// This notably needs to be as fast as possible, and find the car faster then the above function (or should do)
	public boolean findCarAroundExistingPoint(BufferedImage liveImage, BufferedImage image, int x, int y, int minX, int maxX, int minY, int maxY)
	{
		// Valid images
		if(liveImage == null || image == null || liveImage.getWidth() == 0 || image.getWidth() == 0)
		{
			parentCarAIMain.log("[Car] findCarAroundExistingPoint() - Error in input images.");
			return false;
		}
		// This should do a similar search to the above, but around X, Y

		// If we are using an X, Y co-ordinate outside our area, we default and note it
		int newX = Utility.getBoundriedInt(x, minX, maxX);
		int newY = Utility.getBoundriedInt(y, minY, maxY);
		if(newX != x || newY != y)
		{
			// Log
			parentCarAIMain.logDetail("[Car] findCarAroundExistingPoint() - Error, X or Y is outside area. Altering to be within chosen area.");
			x = newX;
			y = newY;
		}

		int xSub = 0, ySub = 0, top = y, left = x, right = x, bottom = y, xCar = 0, yCar = 0, matches = 0;
		// Performance tuning should check here.
		// This is the area size (2 * countPixelAreaSize * 2 * countPixelAreaSize total pixels) around a full pixel we look to see if there is enough to consider the car being part of the area
		int countPixelAreaSize = 3;
		// This is the area we use to average and find the centre of the car location. We find quickly a tiny location which should be it, and average all nearby points of relevant "fullness" to get the general centre
		// Experiment with this number if it isn't entirely accurate all the time.
		int averageLocationAreaSize = 17;	
		// This should be set to the "minimum" area that can be considered part of the car. For performance, set lower.
		int minimumValidPoints = 20;		
		int highestmatches = 0;

		parentCarAIMain.logDetail("[Car] findCarAroundExistingPoint. Looking for car at: X and Y: (" + x + "," + y + "). Full: [" + FULL_REGION + "] Empty: [" + EMPTY_REGION + "] Min: [" + minimumValidPoints + "]");

		// Loop initial pixel only...
		if(image.getRGB(x, y) == MATCH_FULL_REGION)
		{
			matches = countPixelsInArea(x, y, countPixelAreaSize, FULL_REGION, image);
			if(matches > highestmatches)
			{
				highestmatches = matches;
				xCar = x;
				yCar = y;
			}
		}
		
		// Loop the image...
		while(highestmatches < minimumValidPoints && (top > minY || left > minX || bottom < maxY || right < maxX))
		{
			// Expand search, by 1 pixel area more - meaning, an area +2 pixels wide
			top = Utility.getBoundriedInt(top - 1, minY, maxY);// IE: Min Y
			left = Utility.getBoundriedInt(left - 1, minX, maxX);// IE: Min X
			bottom = Utility.getBoundriedInt(bottom + 1, minY, maxY);// IE: Max Y
			right = Utility.getBoundriedInt(right + 1, minX, maxX);// IE: Max X

			// Search along the "top" (nb: we do duplicate the corners, but nevermind) if not the top line (we ignore the top line, won't find the car there, saves CPU if a large area search)
			if(top > minY)
			{
				for(xSub = left; xSub < right; xSub++)
				{
					// Get the RGB values here. If full, we will search around
					if(image.getRGB(xSub, top) == MATCH_FULL_REGION)
					{				
						matches = countPixelsInArea(xSub, top, countPixelAreaSize, MATCH_FULL_REGION, image);
						if(matches > highestmatches)
						{
							highestmatches = matches;
							xCar = xSub;
							yCar = top;
						}
					}
				}
			}
			// "Bottom"
			if(bottom < maxY)
			{
				for(xSub = left; xSub < right; xSub++)
				{
					// Get the RGB values here. If full, we will search around
					if(image.getRGB(xSub, bottom) == MATCH_FULL_REGION)
					{				
						matches = countPixelsInArea(xSub, bottom, countPixelAreaSize, MATCH_FULL_REGION, image);
						if(matches > highestmatches)
						{
							highestmatches = matches;
							xCar = xSub;
							yCar = bottom;
						}
					}
				}
			}
			// "Left"
			if(left > minX)
			{
				for(ySub = top; ySub < bottom; ySub++)
				{
					// Get the RGB values here. If full, we will search around
					if(image.getRGB(left, ySub) == MATCH_FULL_REGION)
					{				
						matches = countPixelsInArea(left, ySub, countPixelAreaSize, MATCH_FULL_REGION, image);
						if(matches > highestmatches)
						{
							highestmatches = matches;
							xCar = left;
							yCar = ySub;
						}
					}
				}
			}
			// "Right"
			if(right < maxX)
			{
				for(ySub = top; ySub < bottom; ySub++)
				{
					// Get the RGB values here. If full, we will search around
					if(image.getRGB(right, ySub) == MATCH_FULL_REGION)
					{				
						matches = countPixelsInArea(right, ySub, countPixelAreaSize, MATCH_FULL_REGION, image);
						if(matches > highestmatches)
						{
							highestmatches = matches;
							xCar = right;
							yCar = ySub;
						}
					}
				}
			}
		}
		// Found something?
		if(highestmatches >= minimumValidPoints)
		{
			// Yes!
			// Find average location now
			Coordinate xyAverage = findAverageColorLocationInArea(xCar, yCar, averageLocationAreaSize, MATCH_FULL_REGION, image);

			// Now we have this car's location find it's facing
			double facing = findCarFacing(xyAverage.x, xyAverage.y, liveImage, image);

			// Errors?
			if(facing < 0)
			{
				// Cannot be negative, we have a problem. We simply don't know the facing!
				parentCarAIMain.logDetail("[Car] findCarAroundExistingPoint. Error finding facing. Using old co-ordinates or previous facing to calculate new facing...");

				// Did we actually move very much?
				if(Math.abs(lastCarCoords.x - xyAverage.x) < 5 && Math.abs(lastCarCoords.y - xyAverage.y) < 5)
				{
					// Use old angle
					facing = lastCarFacing;
				}
				else
				{
					// New facing, since we have moved far enough to calculate a new angle
					// This can cause large errors, so report this is being used
					parentCarAIMain.log("[Car] findCarAroundExistingPoint. Didn't find front of car by colour. Searching for facing by using last known car coordinates.");
					facing = Utility.calcAngle(lastCarCoords.x, lastCarCoords.y, xyAverage.x, xyAverage.y);
				}
				if(facing < 0)
				{
					parentCarAIMain.log("[Car] findCarAroundExistingPoint. Error finding facing again somehow. Returning no facing!");
					return false;
				}
				// Draw the front point (a bit in front of the car)
				// 10 pixels in front
				// Angle is based on coordinates from top left going to bottom right
				camera.drawBlock(Utility.getProjectedPoint(lastCarCoords, xyAverage, 10, minX, maxX, minY, maxY), image, Color.BLUE);
			}

			// Set our new "past" coordinates
			lastCarCoords.x = xyAverage.x;
			lastCarCoords.y = xyAverage.y;
			lastCarFacing = facing;

			// Got a centre location for our car, hopefully. Debug all of this.
			parentCarAIMain.logDetail("[Car] findCarAroundExistingPoint. Found a car at: (" + xCar + "," + yCar + "). Highest matches [" + highestmatches + "] minimumValidPoints [" + minimumValidPoints + "]. Average, and thus CarCentre is at (" + xyAverage.x + "," + xyAverage.y + ") Angle of car: " + facing);
			
			// Log Coordinates
			parentCarAIMain.logCoords("[Car] Car is at: (" + xCar + "," + yCar + ") facing [" + facing + "]");

			// Draw the car
			camera.drawBlock(xyAverage.x, xyAverage.y, image);
			
			return true;
		}

		// No :(
		parentCarAIMain.log("[Car] findCarAroundExistingPoint. No car found! Highest matches [" + highestmatches + "] minimumValidPoints [" + minimumValidPoints + "]");			

		// Debug image
		parentCarAIMain.staticCamImg.image.setImage(image);
		
		return false;
	}
	
	
	/*************************
		PRIVATE FUNCTIONS 
	*************************/
	
	// Using the car location x, y, work out which way the car is facing.
	// Pass in only a true image
	// Returns 0-360.0000, or -1 on error
	// Outputs a square of the "front" on debugImage
	private double findCarFacing(int x, int y, BufferedImage image, BufferedImage debugImage) {
		// Angle to return
		double angle = -1.0;
 
		// x, y should be the car's centre
		// We search for "red" (channel 0).
		Coordinate xyAverage = findAverageThresholdColorLocationInArea(x, y, 13, 0, 180, 255, image, debugImage);

		int frontX = xyAverage.x;
		int frontY = xyAverage.y;

		// Error checking
		if(frontX == 0 || frontY == 0)
		{
			// Error!
			parentCarAIMain.log("[Car] findCarFacing. ERROR. Given X, Y (" + x + "," + y + "), front X,Y (" + frontX + "," + frontY + ").");

			// Return -1.0
			return angle;
		}

		// Debug
		camera.drawBlock(frontX, frontY, debugImage, Color.BLUE);

		// Get the angle
		angle = Utility.calcAngle(x, y, frontX, frontY);

		// Debug
		parentCarAIMain.logDetail("[Car] findCarFacing. Given X, Y (" + x + "," + y + "), front X,Y (" + frontX + "," + frontY + "). Angle [" + angle + "]");

		// Return this angle		
		return angle;		
	}

	
	// Count then number of pixels matching colour "colorMatch".
	// size should be distance from the initial pixel (in a square) to search
	private int countPixelsInArea(int x, int y, int size, int colorMatch, BufferedImage image) {
		// Can be 0 matches of course
		int matches = 0;
		// Search around this region in a smallish space
		int searchStartX = Utility.getBoundriedInt(x - size, trackMinX, trackMaxX);
		int searchStartY = Utility.getBoundriedInt(y - size, trackMinY, trackMaxY);
		int searchEndX = Utility.getBoundriedInt(x + size, trackMinX, trackMaxX);
		int searchEndY = Utility.getBoundriedInt(y + size, trackMinY, trackMaxY);
	
		// Loop image
		for (int xSub = searchStartX; xSub < searchEndX; xSub++)
		{
			for (int ySub = searchStartY; ySub < searchEndY; ySub++)
			{
				// Need a lot of matches
				if(image.getRGB(xSub, ySub) == colorMatch)
				{
					matches++;
				}
			}
		}
		return matches;
	}

	// Average the location of pixels matching colorMatch around x, y +/- size distance away on the inputted image.
	// Returns an array, 0 is X, 1 is Y for now
	private Coordinate findAverageColorLocationInArea(int x, int y, int size, int colorMatch, BufferedImage image) {
		Coordinate returnCoords = new Coordinate();
	
		// Variables for loop
		int amountOfPoints = 0, xTotal = 0, yTotal = 0;

		// ...using this point...
		// ...loop around a 30x30 area around the car (this covers any angle and direction, actual car image is roughly 20-25 "long") 
		// * The point initially found is usually on the edge of the car area, due to it using a smaller search area above
		int searchStartX = Utility.getBoundriedInt(x - size, trackMinX, trackMaxX);
		int searchEndX = Utility.getBoundriedInt(x + size, trackMinX, trackMaxX);
		int searchStartY = Utility.getBoundriedInt(y - size, trackMinY, trackMaxY);
		int searchEndY = Utility.getBoundriedInt(y + size, trackMinY, trackMaxY);

		// Loop
		for (int xSub = searchStartX; xSub < searchEndX; xSub++)
		{
			for (int ySub = searchStartY; ySub < searchEndY; ySub++)
			{				
				// Average the location of the points found to get the "car centre" by adding on the x and y co-ordinates
				if(image.getRGB(xSub, ySub) == colorMatch)
				{
					amountOfPoints++;
					xTotal += xSub;
					yTotal += ySub;
				}
			}
		}	
		// Average...
		if(amountOfPoints > 0)
		{
			returnCoords.x = xTotal / amountOfPoints;
			returnCoords.y = yTotal / amountOfPoints;
		}
		// DEBUG
		/*
		// Colour corners of the search area now we're done searching
		drawBlock(x, y, image, Color.PINK);// Centre of search area
		drawBlock(searchStartX, searchStartY, image, Color.WHITE);// Top left
		drawBlock(searchEndX, searchStartY, image, Color.GRAY);// Top right
		drawBlock(searchStartX, searchEndY, image, Color.BLUE);// Bottom left
		drawBlock(searchEndX, searchEndY, image, Color.YELLOW);// Bottom right
		// Mark final car loaction on image
		drawBlock(returnArray[0], returnArray[1], image);
		//*/
		
		// Return the car's position
		return returnCoords;
	}

	// Average the location of pixels matching thresholdTop/Bottom of type rgb around x, y +/- size distance away on the inputted image.
	// Returns an array, 0 is X, 1 is Y for now
	// rgbChoice - red = 0, green = 1, blue = 2
	private Coordinate findAverageThresholdColorLocationInArea(int x, int y, int size, int rgbChoice, int thresholdBottom, int thresholdTop, BufferedImage image, BufferedImage debugImage) {
		Coordinate returnCoords = new Coordinate();

		// Variables for loop
		int xAmountOfPoints = 0, yAmountOfPoints = 0, xTotal = 0, yTotal = 0, rgb = 0, r = 0, g = 0, b = 0, choiceColor = 0;
		boolean chosen = false;

		// ...using this point...
		// ...loop around a 30x30 area around the car (this covers any angle and direction, actual car image is roughly 20-25 "long") 
		// * The point initially found is usually on the edge of the car area, due to it using a smaller search area above
		int searchStartX = Utility.getBoundriedInt(x - size, trackMinX, trackMaxX);
		int searchEndX = Utility.getBoundriedInt(x + size, trackMinX, trackMaxX);
		int searchStartY = Utility.getBoundriedInt(y - size, trackMinY, trackMaxY);
		int searchEndY = Utility.getBoundriedInt(y + size, trackMinY, trackMaxY);

		//System.out.println("findAverageThresholdColorLocationInArea (" + searchStartX + "," + searchStartY + ") to (" + searchEndX + "," + searchEndY + "). thresholdBottom: " + thresholdBottom + ". thresholdTop" + thresholdTop + ".");	

		// Loop
		for (int xSub = searchStartX; xSub < searchEndX; xSub++)
		{
			for (int ySub = searchStartY; ySub < searchEndY; ySub++)
			{
				// Within a threshold of colour
				rgb = image.getRGB(xSub, ySub);
				r = (rgb & 0xFF0000) >> 16;
				g = (rgb & 0x00FF00) >> 8;
				b = (rgb & 0x0000FF);
				chosen = false;
				if(rgbChoice == 0)
				{
					// Red
					if(r >= thresholdBottom && r <= thresholdTop && g < 100 && b < 100)
					{
						// Choose
						chosen = true;
					}
				}
				else if(rgbChoice == 1)
				{
					// Green
					if(g >= thresholdBottom && g <= thresholdTop && r < 100 && b < 100)
					{
						// Choose
						chosen = true;
					}
				}
				else //if(rgbChoice == 2)
				{
					// Blue
					if(b >= thresholdBottom && b <= thresholdTop && r < 100 && g < 100)
					{
						// Choose
						chosen = true;
					}		
				}

				if(chosen == true)
				{
					// Average the location of the points found to get the "car centre" by adding on the x and y co-ordinates
					xAmountOfPoints++;
					xTotal += xSub;
					yAmountOfPoints++;
					yTotal += ySub;

					// Debug
					//debugImage.setRGB(xSub, ySub, Color.BLUE.getRGB());
					//System.out.println("Channel: [" + rgbChoice + "] (" + xSub + "," + ySub + ") being searched. Total RGB chosen: " + choiceColor);
				}
			}
		}
		// Average to get result
		if(xAmountOfPoints > 0 && yAmountOfPoints > 0)
		{
			returnCoords.x = xTotal / xAmountOfPoints;
			returnCoords.y = yTotal / yAmountOfPoints;
			//System.out.println("findAverageThresholdColorLocationInArea Average: (" + returnArray[0] + "," + returnArray[1] + ")");
		}
		// DEBUG
		/*
		// Colour corners of the search area now we're done searching
		drawBlock(x, y, image, Color.PINK);// Centre of search area
		drawBlock(searchStartX, searchStartY, image, Color.WHITE);// Top left
		drawBlock(searchEndX, searchStartY, image, Color.GRAY);// Top right
		drawBlock(searchStartX, searchEndY, image, Color.BLUE);// Bottom left
		drawBlock(searchEndX, searchEndY, image, Color.YELLOW);// Bottom right
		// Mark final car loaction on image
		drawBlock(returnArray[0], returnArray[1], image, Color.PINK);
		// END DEBUG
		//*/
		
		// Return the car's position
		return returnCoords;
	}
}