/**
 * CarAIMain.java
 *
 * Entry class - contains the essential void main() of the program, and will start the program running.
 * Creates the GUI and loads the initial settings.
 *
 * Makes use of code created/modified by Chris Carter and Scott Culcheth (see associated individual classes for details)
 *
 * Created by: Chris Carter (Original file: Simulation.java)
 * Modified by: Andrew Armstrong
 * Last Modified: 04.05.08
 *
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import ppv.*;
import java.lang.*;
// For logging
import java.io.*;
// For date stuff
import java.util.*;

public class CarAIMain extends JFrame implements WindowListener{

	// Engine - this controls the program, and deals with GUI inputs
	Engine engine;
	
	// Need SimDraw still since it acts like a GUI element
	SimDraw simDraw;

	// Log file
	FileOutputStream logFile;
	PrintStream logStream;
	boolean saveLog = true; // True creates log file.
	boolean detInfo = false;// Detailed logging info toggle. If true, the functions for detailed info work.

	// Image processing controls
	// Image thresholding	
	JSlider thresholdRedBarSlider;
	JLabel thresholdRedBarLabelValue;
	JSlider thresholdGreenBarSlider;
	JLabel thresholdGreenBarLabelValue;
	JSlider thresholdBlueBarSlider;
	JLabel thresholdBlueBarLabelValue;
	JButton thresholdButton = new JButton("Threshold Image");
	// Background subtracter
	JSlider backgroundSubtracterSlider;
	JLabel backgroundSubtracterSliderValue;
	JButton backgroundSubtracterButton = new JButton("Background Subtraction");
	JButton saveReferenceButton = new JButton("Save Reference Image");
	// Frame differencing (button put with background subtracter)
	JButton frameDifferencingButton = new JButton("Frame Difference");

	// Panel has a set of test buttons on it (there was an old "Debug Button" but test button makes more sense).
	JButton testButton1 = new JButton("Test Button 1");
	JButton testButton2 = new JButton("Test Button 2");
	JButton testButton3 = new JButton("Test Button 3");
	JButton testButton4 = new JButton("Test Button 4");
	JButton testButtonFollowingStart = new JButton("Test Following Start");
	JButton testButtonFollowingEnd = new JButton("Test Following End");
	
	// Declare buttons and text feedback globally so they can be accessed by other classes
	JButton startButton = new JButton("Start Simulation");
	JButton stopButton = new JButton("Stop Simulation");
	JButton resetButton = new JButton("Reset");
	JButton camButton = new JButton ("Turn On Camera");

	// AA: 3 new buttons. Save/Load image.
	JButton saveLiveImageButton = new JButton("Save Live PNG Image");
	JButton saveStaticImageButton = new JButton("Save Static PNG Image");
	JButton loadImageButton = new JButton ("Load Image");
	
	// AA: AI control buttons (in their own box)
	JButton setupAIButton = new JButton("Setup AI");
	JButton startAIButton = new JButton("Start AI");
	JButton pauseResumeAIButton = new JButton("Pause AI"); // Also "Resume AI"
	JButton stopAIButton = new JButton("STOP AI");
	JButton displayPastPointsAIButton = new JButton("Display Past Points");
	JButton incrementStartingPointAIButton = new JButton("Increment Starting point");
	
	// Speed slider
	JSlider clockSlider = new JSlider(0, 50, 1);
	JTextField clockText = new JTextField(2);

	// Option checkboxes
	JCheckBox driveCarCB = new JCheckBox("Simulation Drives Car", true);
	JCheckBox stopCrashCB = new JCheckBox("Stop Simulation on collision", true);
	JCheckBox detInfoCB = new JCheckBox("Detailed Log Info", detInfo);
	JCheckBox saveLogCB = new JCheckBox("Save Log to file", saveLog);
	JCheckBox crashRecoveryCB = new JCheckBox("Crash Recovery Only", false);

	// Definition of info and coordinates text debug boxes
	// - Privatised so cannot be messed with except for in the log() functions.
	private JTextArea infoPanel = new JTextArea(8,40);
	private JScrollPane infoScroll = new JScrollPane(infoPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
	private JTextArea coordPanel = new JTextArea(3,40);
	private JScrollPane coordScroll = new JScrollPane(coordPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

	// Camera image - the live image
	CameraImage camImg;
	// Second camera image for things like drawing waypoints, track lines, frame differencing, and debugging.
	CameraImageStatic staticCamImg;

	// I've researched this, and it is generally only used for sorting deployments - if it is unaltered, old .class files can work with newer ones.
	// Change it, and everything needs recompiling. Also some issues with extend support? I've decided to leave it alone :) - http://today.java.net/pub/a/today/2006/03/09/serialVersionUIDSQ8.html
	private final static long serialVersionUID = 1L;

	// Constructor
	public CarAIMain() {
		// Set the window text
		super ("Car Racing AI by Andrew Armstrong");
		// Full screen the window
		setExtendedState(Frame.MAXIMIZED_BOTH);
		// Set the background colour to "gray" (or grey really)
		setBackground(Color.gray);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

		// *******************************
		// Layout creation
		// *******************************
		// Layout is now on the basis of LHS: Camera / Simulation windows, RHS: Controls
		// We use a grid to layout the controls equally into separate JPanels. SimDraw is a JPanel itself too!
/*
	mainPanel: contains left and right panels (left *should* be at least big enough to fit the camera images or simulation, whichever is biggest)
	
		1	2

	leftPanel: contains the simulation and the camera images (or placeholders)
		
		1
		2
		...
		
	rightPanel: contains X sets of JPanel's for the controls, 1 column
	
		1
		2
		...
		...
*/
		// *************************
		// Create main panel
		// *************************
		Container simDisplay = getContentPane();
		JPanel mainPanel = new JPanel(); // Everything is contained in this mainPanel
		mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.X_AXIS));
		mainPanel.setBackground(Color.gray);

		// *************************
		// AA: Panel for the 3 "camera's" on the left
		// *************************
		JPanel leftPanel = new JPanel();
		leftPanel.setLayout(new BoxLayout(leftPanel, BoxLayout.Y_AXIS));
		leftPanel.setBackground(Color.gray);
		leftPanel.setMinimumSize(new Dimension(700, 0)); // 1280 / 2 = 640, but need to round to 700. It tends to make right hand too small otherwise.
		mainPanel.add(leftPanel);

		// *************************
		// Add panel for the other controls on the right, 5 panels
		// *************************
		JPanel rightPanel = new JPanel();
		rightPanel.setLayout(new BoxLayout(rightPanel, BoxLayout.Y_AXIS));
		rightPanel.setBackground(Color.gray);
		rightPanel.setPreferredSize(new Dimension(580, 0)); // 1280 / 2 = 640, but take off 60 so 580. It tends to make right hand too small otherwise.
		mainPanel.add(rightPanel);

		// *************************
		JLabel simDrawLabel = new JLabel("Simulation (Currently unused with AI)");
		simDrawLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		simDrawLabel.setForeground(Color.black);
		leftPanel.add(simDrawLabel);
		// Draw track layout and car starting position, and prepare simulation.
		simDraw = new SimDraw(this);
		// Grey background to cover whole window (so that white walls are clearer).
		simDraw.setBackground(Color.gray);
		simDraw.setPreferredSize(new Dimension(640, 330));
		// Add the image panel drawn by SimDraw to LHS frame
		leftPanel.add(simDraw);

		// *************************
		// Cameras
		// *************************

		// Camera panel is to put the camera images next to each other
		JPanel cameraPanel = new JPanel();
		cameraPanel.setLayout(new BoxLayout(cameraPanel, BoxLayout.LINE_AXIS));
		cameraPanel.setBackground(Color.gray);

		// Camera 1 panel (text above the image, left hand side)
		JPanel camera1Panel = new JPanel();
		camera1Panel.setLayout(new BoxLayout(camera1Panel, BoxLayout.PAGE_AXIS));
		camera1Panel.setBackground(Color.lightGray);
		camera1Panel.setPreferredSize(new Dimension(340, 300));

		JLabel primaryCamLabel = new JLabel("Primary Camera (Live)");
		primaryCamLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		primaryCamLabel.setForeground(Color.black);
		camera1Panel.add(primaryCamLabel);
		camImg = new CameraImage();
		camImg.setBackground(Color.gray);
		camera1Panel.add(camImg);

		// Live camera buttons added here
		JPanel camera1ButtonPanel = new JPanel();
		camera1ButtonPanel.setLayout(new BoxLayout(camera1ButtonPanel, BoxLayout.PAGE_AXIS));
		camera1ButtonPanel.setBackground(Color.lightGray);
		camera1ButtonPanel.setPreferredSize(new Dimension(0, 100));
		
		camButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		camera1ButtonPanel.add(camButton);
		saveLiveImageButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		camera1ButtonPanel.add(saveLiveImageButton);
		camera1Panel.add(camera1ButtonPanel);
		
		// Add camera 1
		cameraPanel.add(camera1Panel);

		// Camera 2 panel (text above the image, right hand side)
		JPanel camera2Panel = new JPanel();
		camera2Panel.setLayout(new BoxLayout(camera2Panel, BoxLayout.PAGE_AXIS));
		camera2Panel.setBackground(Color.lightGray);
		camera2Panel.setPreferredSize(new Dimension(340, 300));

		// Add static image for other functions
		JLabel staticCamImgLabel = new JLabel("Secondary Camera (Debugging/Static Image)");
		staticCamImgLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		staticCamImgLabel.setForeground(Color.black);
		camera2Panel.add(staticCamImgLabel);
		staticCamImg = new CameraImageStatic();
		staticCamImg.setBackground(Color.gray);
		camera2Panel.add(staticCamImg);
		saveReferenceButton.setAlignmentX(Component.CENTER_ALIGNMENT);

		// Static camera buttons
		JPanel camera2ButtonPanel = new JPanel();
		camera2ButtonPanel.setLayout(new BoxLayout(camera2ButtonPanel, BoxLayout.PAGE_AXIS));
		camera2ButtonPanel.setBackground(Color.lightGray);
		camera2ButtonPanel.setPreferredSize(new Dimension(0, 100));
		
		saveReferenceButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		camera2ButtonPanel.add(saveReferenceButton);
		saveStaticImageButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		camera2ButtonPanel.add(saveStaticImageButton);
		// Load image
		loadImageButton.setAlignmentX(Component.CENTER_ALIGNMENT);
		camera2ButtonPanel.add(loadImageButton);
		camera2Panel.add(camera2ButtonPanel);
		
		// Add camera 2
		cameraPanel.add(camera2Panel);
		
		// Add camera panel
		leftPanel.add(cameraPanel);
		
		// *************************
		// Start engine to run everything. This has action events added later for the GUI controls. The main thing it requires is drawim and camImg to be created before being called.
		engine = new Engine(this);
		// *************************
		
		
		// *************************
		// Camera image controls
		JLabel camControlsLabel = new JLabel("Image controls");
		camControlsLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		camControlsLabel.setForeground(Color.black);
		leftPanel.add(camControlsLabel);

		// Image thresholder 
		// * 3 bars for R, G and B values (being made up of "Label" "Bar" "Label"), with a button on the right to "use" it

		// More boxes, wheee!
		JPanel thresholdPanel = new JPanel();
		thresholdPanel.setLayout(new BoxLayout(thresholdPanel, BoxLayout.X_AXIS));
		thresholdPanel.setBackground(Color.lightGray);
		JPanel thresholdBarPanel = new JPanel();// Contains the 3 bar panels
		thresholdBarPanel.setLayout(new BoxLayout(thresholdBarPanel, BoxLayout.Y_AXIS));
		thresholdBarPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		thresholdBarPanel.setBackground(Color.lightGray);
		
		// RED
		JPanel thresholdRedBarPanel = new JPanel();
		thresholdRedBarPanel.setLayout(new BoxLayout(thresholdRedBarPanel, BoxLayout.X_AXIS));
		thresholdRedBarPanel.setBackground(Color.lightGray);
		// Slider
		thresholdRedBarSlider = new JSlider(0,256, engine.camera.imageThresholder.getRHi());
		thresholdRedBarSlider.setMajorTickSpacing(16);
		thresholdRedBarSlider.setMinorTickSpacing(4);
		thresholdRedBarSlider.setPaintTicks(true);
		thresholdRedBarSlider.setForeground(Color.black);
		thresholdRedBarSlider.setBackground(Color.lightGray);
		//thresholdRedBarSlider.setMaximumSize(new Dimension(350, 30));
		// Description label
		JLabel thresholdRedBarLabel = new JLabel("  RED:");
		thresholdRedBarLabel.setFont(new Font("Monospaced", Font.PLAIN, 14));// A cludge to stop me spending hours on layouts.
		thresholdRedBarLabel.setLabelFor(thresholdRedBarSlider);
		// Add slider to panel
		thresholdRedBarPanel.add(thresholdRedBarLabel);
		thresholdRedBarPanel.add(thresholdRedBarSlider);
		// Label for value (convert the initial value to a string too)
		thresholdRedBarLabelValue = new JLabel(Integer.toString(engine.camera.imageThresholder.getRHi()));
		thresholdRedBarLabelValue.setPreferredSize(new Dimension(24, 0));
		thresholdRedBarPanel.add(thresholdRedBarLabelValue);		
		// Add red to thresholdBarPanel
		thresholdBarPanel.add(thresholdRedBarPanel);

		// GREEN
		JPanel thresholdGreenBarPanel = new JPanel();
		thresholdGreenBarPanel.setLayout(new BoxLayout(thresholdGreenBarPanel, BoxLayout.X_AXIS));
		thresholdGreenBarPanel.setBackground(Color.lightGray);
		// Slider
		thresholdGreenBarSlider = new JSlider(0,256, engine.camera.imageThresholder.getGHi());
		thresholdGreenBarSlider.setMajorTickSpacing(16);
		thresholdGreenBarSlider.setMinorTickSpacing(4);
		thresholdGreenBarSlider.setPaintTicks(true);	
		thresholdGreenBarSlider.setForeground(Color.black);
		thresholdGreenBarSlider.setBackground(Color.lightGray);
		// Description label
		JLabel thresholdGreenBarLabel = new JLabel("GREEN:");
		thresholdGreenBarLabel.setFont(new Font("Monospaced", Font.PLAIN, 14));
		thresholdGreenBarLabel.setLabelFor(thresholdGreenBarSlider);
		// Add slider to panel
		thresholdGreenBarPanel.add(thresholdGreenBarLabel);
		thresholdGreenBarPanel.add(thresholdGreenBarSlider);
		// Label for value (convert the initial value to a string too)
		thresholdGreenBarLabelValue = new JLabel(Integer.toString(engine.camera.imageThresholder.getGHi()));
		thresholdGreenBarLabelValue.setPreferredSize(new Dimension(24, 0));
		thresholdGreenBarPanel.add(thresholdGreenBarLabelValue);
		// Add green to thresholdBarPanel
		thresholdBarPanel.add(thresholdGreenBarPanel);

		// BLUE
		JPanel thresholdBlueBarPanel = new JPanel();
		thresholdBlueBarPanel.setLayout(new BoxLayout(thresholdBlueBarPanel, BoxLayout.X_AXIS));
		thresholdBlueBarPanel.setBackground(Color.lightGray);
		// Slider
		thresholdBlueBarSlider = new JSlider(0,256, engine.camera.imageThresholder.getBHi());
		thresholdBlueBarSlider.setMajorTickSpacing(16);
		thresholdBlueBarSlider.setMinorTickSpacing(4);
		thresholdBlueBarSlider.setPaintTicks(true);
		thresholdBlueBarSlider.setForeground(Color.black);
		thresholdBlueBarSlider.setBackground(Color.lightGray);
		// Description label
		JLabel thresholdBlueBarLabel = new JLabel(" BLUE:");
		thresholdBlueBarLabel.setFont(new Font("Monospaced", Font.PLAIN, 14));
		thresholdBlueBarLabel.setLabelFor(thresholdBlueBarSlider);
		// Add slider to panel
		thresholdBlueBarPanel.add(thresholdBlueBarLabel);
		thresholdBlueBarPanel.add(thresholdBlueBarSlider);		
		// Label for value (convert the initial value to a string too)
		thresholdBlueBarLabelValue = new JLabel(Integer.toString(engine.camera.imageThresholder.getBHi()));
		thresholdBlueBarLabelValue.setPreferredSize(new Dimension(24, 0));
		thresholdBlueBarPanel.add(thresholdBlueBarLabelValue);		
		// Add blue to thresholdBarPanel
		thresholdBarPanel.add(thresholdBlueBarPanel);

		// Add threshold bars to the threshold panel
		thresholdPanel.add(thresholdBarPanel);

		// Add the button for the "test this" bit
		thresholdPanel.add(thresholdButton);		
		
		// Add the thresholder stuff to the left panel
		leftPanel.add(thresholdPanel);

		// Background subtraction
		// * 1 bar for this, and a button.
		JPanel backgroundSubtracterPanel = new JPanel();
		backgroundSubtracterPanel.setLayout(new BoxLayout(backgroundSubtracterPanel, BoxLayout.Y_AXIS));
		backgroundSubtracterPanel.setBackground(Color.lightGray);

		// Contains the slider
		JPanel backgroundSubtracterSliderPanel = new JPanel();
		backgroundSubtracterSliderPanel.setLayout(new BoxLayout(backgroundSubtracterSliderPanel, BoxLayout.X_AXIS));
		backgroundSubtracterSliderPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		backgroundSubtracterSliderPanel.setBackground(Color.lightGray);

		// Slider
		backgroundSubtracterSlider = new JSlider(0,256, engine.camera.backgroundSubtracter.getThreshold());
		backgroundSubtracterSlider.setMajorTickSpacing(16);
		backgroundSubtracterSlider.setMinorTickSpacing(4);
		backgroundSubtracterSlider.setPaintTicks(true);
		backgroundSubtracterSlider.setForeground(Color.black);
		backgroundSubtracterSlider.setBackground(Color.lightGray);
		// Description label
		JLabel backgroundSubtracterSliderLabel = new JLabel("Background Subtraction:");
		backgroundSubtracterSliderLabel.setFont(new Font("Monospaced", Font.PLAIN, 14));
		backgroundSubtracterSliderLabel.setLabelFor(backgroundSubtracterSlider);
		// Add slider to panel
		backgroundSubtracterSliderPanel.add(backgroundSubtracterSliderLabel);
		backgroundSubtracterSliderPanel.add(backgroundSubtracterSlider);		
		// Label for value (convert the initial value to a string too)
		backgroundSubtracterSliderValue = new JLabel(Integer.toString(engine.camera.backgroundSubtracter.getThreshold()));
		backgroundSubtracterSliderValue.setPreferredSize(new Dimension(24, 0));
		backgroundSubtracterSliderPanel.add(backgroundSubtracterSliderValue);		
		// Add this to backgroundSubtractionPanel
		backgroundSubtracterPanel.add(backgroundSubtracterSliderPanel);
		// Add buttons to backgroundSubtractionPanel
		JPanel backgroundSubtracterButtonPanel = new JPanel();
		backgroundSubtracterButtonPanel.setLayout(new BoxLayout(backgroundSubtracterButtonPanel, BoxLayout.X_AXIS));
		// Add buttons
		backgroundSubtracterButtonPanel.add(backgroundSubtracterButton);
		backgroundSubtracterButtonPanel.add(frameDifferencingButton);
		// Add to BGsub panel
		backgroundSubtracterPanel.add(backgroundSubtracterButtonPanel);
		
		// Add to left panel
		leftPanel.add(backgroundSubtracterPanel);
		
		// *************************
		// Now do the panels for the right hand side
		// *************************
		
		// Program controls label
		JLabel simulationControlPanelLabel = new JLabel("Simulation Controls");
		simulationControlPanelLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		simulationControlPanelLabel.setForeground(Color.black);
		rightPanel.add(simulationControlPanelLabel);

		// RHS panel for the simulation control buttons + label
		JPanel simulationControlPanel = new JPanel();
		simulationControlPanel.setLayout(new FlowLayout());
		simulationControlPanel.setBackground(Color.lightGray);

		// Add buttons, options and controls to correct panels
		simulationControlPanel.add(startButton);
		simulationControlPanel.add(stopButton);
		simulationControlPanel.add(resetButton);
		
		// Add button pane to the RHS
		rightPanel.add(simulationControlPanel);
		
		// RHS Panel for the speed control - a Clock Rate slider
		JPanel clockPanel = new JPanel();
		clockPanel.setBackground(Color.lightGray);
		// Size of it
		clockSlider.setMajorTickSpacing(10);
		clockSlider.setMinorTickSpacing(2);
		clockSlider.setPaintTicks(true);
		clockSlider.setBackground(Color.lightGray);
		clockSlider.setForeground(Color.gray);
		// Label for the clock speed
		JLabel clockLabel = new JLabel("Sim Speed: ");
		clockLabel.setLabelFor(clockSlider);
		clockLabel.setForeground(Color.black);

		// Add both of these to the clock panel
		clockPanel.add(clockLabel);
		clockPanel.add(clockSlider);

		// Text box displaying current value of clock rate slider
		clockText.setEditable(false);
		clockText.setFont(new Font("f1", Font.ITALIC, 18));
		clockText.setForeground(Color.black);
		clockText.setOpaque(true);
		clockText.setBackground(Color.lightGray);
		int clockRate_CB = clockSlider.getValue();
		Integer clockRate = new Integer( clockRate_CB );
		clockText.setText(clockRate.toString());
		// Add this to the panel
		clockPanel.add(clockText);

		// Add to RHS panel
		rightPanel.add(clockPanel);

		// RHS panel for AI controls
		JLabel buttonAIPanelLabel = new JLabel("AI and Driving Controls");
		buttonAIPanelLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		buttonAIPanelLabel.setForeground(Color.black);
		rightPanel.add(buttonAIPanelLabel);
		// The panel
		JPanel buttonAIPanel = new JPanel();
		buttonAIPanel.setLayout(new FlowLayout());
		buttonAIPanel.setBackground(Color.lightGray);
		// Add AI buttons
		buttonAIPanel.add(setupAIButton);
		buttonAIPanel.add(startAIButton);
		buttonAIPanel.add(pauseResumeAIButton);
		buttonAIPanel.add(stopAIButton);
		buttonAIPanel.add(Box.createGlue());
		buttonAIPanel.add(displayPastPointsAIButton);
		buttonAIPanel.add(incrementStartingPointAIButton);
		// Add AI panel to pane
		rightPanel.add(buttonAIPanel);

		// Options panel label
		JLabel checkButtonLabel = new JLabel("Run Options");
		checkButtonLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		checkButtonLabel.setForeground(Color.black);	
		rightPanel.add(checkButtonLabel);

		// The checkboxes		
		JPanel checkButtonPanel = new JPanel();
		checkButtonPanel.setBackground(Color.lightGray);//gray);
		checkButtonPanel.setLayout(new GridLayout(3, 2));// "X rows, 2 columns"

		driveCarCB.setBackground(Color.lightGray);
		driveCarCB.setForeground(Color.black);
		driveCarCB.addItemListener(engine);

		stopCrashCB.setBackground(Color.lightGray);
		stopCrashCB.setForeground(Color.black);
		stopCrashCB.addItemListener(engine);

		detInfoCB.setBackground(Color.lightGray);
		detInfoCB.setForeground(Color.black);
		detInfoCB.addItemListener(engine);

		saveLogCB.setBackground(Color.lightGray);
		saveLogCB.setForeground(Color.black);
		saveLogCB.addItemListener(engine);

		crashRecoveryCB.setBackground(Color.lightGray);
		crashRecoveryCB.setForeground(Color.black);
		crashRecoveryCB.addItemListener(engine);
	
		checkButtonPanel.add(driveCarCB);
		checkButtonPanel.add(stopCrashCB);
		checkButtonPanel.add(detInfoCB);
		checkButtonPanel.add(saveLogCB);
		checkButtonPanel.add(crashRecoveryCB);

		// Add check button panel to RHS
		rightPanel.add(checkButtonPanel);

		// This is the A-B-C-D diagram of the car layout		
		JLabel keyLabel = new JLabel("Simulation log key (diagram of car)");
		keyLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		keyLabel.setForeground(Color.black);
		rightPanel.add(keyLabel);
		ImageIcon key = new ImageIcon("carKey.png");
		JLabel keyLbl = new JLabel(key);
		keyLbl.setAlignmentX(Component.CENTER_ALIGNMENT);

		// Add to RHS
		rightPanel.add(keyLbl);
		
		// Add image depicting the Parallel port monitor
		JLabel ppMonitorLabel = new JLabel("Parallel Port Monitor");
		ppMonitorLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		ppMonitorLabel.setForeground(Color.black);	
		rightPanel.add(ppMonitorLabel);
		
		PortViewerPanel pPVPanel = new PortViewerPanel();
		//pPVPanel.setAlignment
		pPVPanel.setBackground(Color.gray);

		// Container for PPV
		JPanel pPVPanelSuper = new JPanel();
		// Center and set height to 144 maximum
		pPVPanelSuper.setMaximumSize(new Dimension(Integer.MAX_VALUE, 144));
		pPVPanelSuper.setAlignmentX(Component.CENTER_ALIGNMENT);
		pPVPanelSuper.add(pPVPanel);

		// Add to RHS
		rightPanel.add(pPVPanelSuper);

		// Panel for test buttons
		JPanel testButtonsPanel = new JPanel();

		// Label
		JLabel testButtonsPanelLabel = new JLabel("Test Buttons");
		testButtonsPanelLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		testButtonsPanelLabel.setForeground(Color.black);	
		rightPanel.add(testButtonsPanelLabel);
		
		// Add buttons
		testButtonsPanel.add(testButton1);
		testButtonsPanel.add(testButton2);
		testButtonsPanel.add(testButton3);
		testButtonsPanel.add(testButton4);
		testButtonsPanel.add(testButtonFollowingStart);
		testButtonsPanel.add(testButtonFollowingEnd);
		// Add panel to right
		rightPanel.add(testButtonsPanel);
		
		// Debug panel; box "Car Track Co-ords" and "Information"
		JPanel debugPanel = new JPanel();
		debugPanel.setLayout(new BoxLayout(debugPanel, BoxLayout.Y_AXIS));//new GridLayout(0,1)); // X rows, 1 columns
		debugPanel.setBackground(Color.gray);
		debugPanel.setPreferredSize(new Dimension(0, 300));

		// Lock text areas and add scrollbars when necessary
		infoPanel.setEditable(false);
		infoPanel.setLineWrap(true);
		coordPanel.setEditable(false);
		coordPanel.setLineWrap(true);

		// Panel for "Coords" (contains the label + textbox)
		JLabel coordLabel = new JLabel("Log: Car Track Coords (pixels)");
		coordLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		coordLabel.setForeground(Color.black);
		JPanel debugCoordPanel = new JPanel();
		debugCoordPanel.setLayout(new BoxLayout(debugCoordPanel, BoxLayout.Y_AXIS));
		debugCoordPanel.setBackground(Color.gray);
		//debugCoordPanel.setMaximumSize(new Dimension(0, 50));
		debugCoordPanel.add(coordLabel);
		debugCoordPanel.add(coordScroll);

		// Panel for "Info" (contains the label + textbox)
		JLabel infoLabel = new JLabel("Log: Debug Information");
		infoLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
		infoLabel.setForeground(Color.black);
		JPanel debugInfoPanel = new JPanel();
		debugInfoPanel.setLayout(new BoxLayout(debugInfoPanel, BoxLayout.Y_AXIS));
		debugInfoPanel.setBackground(Color.gray);
		debugInfoPanel.setMinimumSize(new Dimension(0, 250));
		debugInfoPanel.add(infoLabel);
		debugInfoPanel.add(infoScroll);

		// Add these elements to the debugPanel
		debugPanel.add(debugCoordPanel);
		debugPanel.add(debugInfoPanel);

		// Add the debug panel to the RHS
		rightPanel.add(debugPanel);

		// *************************
		// Control listeners
		// *************************

		// Add ActionListeners to buttons for the Engine to sort their actions
		startButton.addActionListener(engine);
		stopButton.addActionListener(engine);
		resetButton.addActionListener(engine);
		camButton.addActionListener(engine);
		saveLiveImageButton.addActionListener(engine);
		saveStaticImageButton.addActionListener(engine);
		loadImageButton.addActionListener(engine);
		// AI ones
		setupAIButton.addActionListener(engine);
		startAIButton.addActionListener(engine);
		pauseResumeAIButton.addActionListener(engine);
		stopAIButton.addActionListener(engine);
		displayPastPointsAIButton.addActionListener(engine);
		incrementStartingPointAIButton.addActionListener(engine);
		// Test ones
		testButton1.addActionListener(engine);
		testButton2.addActionListener(engine);
		testButton3.addActionListener(engine);
		testButton4.addActionListener(engine);
		testButtonFollowingStart.addActionListener(engine);
		testButtonFollowingEnd.addActionListener(engine);

		// Click slider
		clockSlider.addChangeListener(engine);
		
		// Image manipulation
		saveReferenceButton.addActionListener(engine);
		
		// Thresholding sliders		
		thresholdRedBarSlider.addChangeListener(engine);
		thresholdGreenBarSlider.addChangeListener(engine);
		thresholdBlueBarSlider.addChangeListener(engine);
		// Threshold button
		thresholdButton.addActionListener(engine);
		
		// BG sub
		backgroundSubtracterSlider.addChangeListener(engine);
		backgroundSubtracterButton.addActionListener(engine);
		// Frame differencing
		frameDifferencingButton.addActionListener(engine);
		
		// Stop and Reset buttons disabled in initial state (as there is nothing to stop/reset!!!)
		stopButton.setEnabled(false);
		resetButton.setEnabled(false);
		// AI buttons
		startAIButton.setEnabled(false);
		pauseResumeAIButton.setEnabled(false);
		stopAIButton.setEnabled(false);
		displayPastPointsAIButton.setEnabled(false);
		incrementStartingPointAIButton.setEnabled(false);

		// Display the main JPanel		
		simDisplay.add(mainPanel);

		setContentPane(simDisplay);
		setVisible(true);

		startButton.requestFocus(true);
		addWindowListener(this);
		// AA: Now resizable until UI is entirely sorted
		setResizable(true);
				
		// Refresh the display
		repaint();
		
		logDetail("[CarAIMain] Complete GUI.");
	}

	// MAIN PROGRAM
	public static void main(String[] args){
	
		// Start a new Simulation
		try{
			CarAIMain programCarAIMain = new CarAIMain();
		}
		catch(Exception e){
			// No logging for this exception - hopefully this will never cause one!
			System.out.println(e);
		}
	}

	/**
	 * Appends a message to the debug console.
	 * @param message the message.
	 */
	public void log(String message) { log(message, false); }
	public void log(String message, boolean replace)
	{
		// For serious errors (ie; infinite loops where the GUI cannot be updated!) uncomment this to provide more functional system out debugging (which is why we don't remove the command line start :) )
		System.out.println(message);

		if(replace == true)
		{
			infoPanel.setText(message);
			// Start a new log file if needed with this
			if(saveLog == true)
			{
				closeLog();
				saveLog = true;
				logToFile(message);
			}
		}
		else 
		{
			infoPanel.append(message + "\n");
			// Jump to this new line - note; this will error if we want to debug before the scroll box is created
			infoScroll.getVerticalScrollBar().setValue(infoScroll.getVerticalScrollBar().getMaximum());
			infoPanel.select(infoPanel.getText().length(), 0);
			// Log to the file if turned on
			logToFile(message);
		}
	}

	// Detailed logging wrapper
	public void logDetail(String message) { logDetail(message, false); }
	public void logDetail(String message, boolean replace)
	{
		// Is the setting on?
		if(detInfo == true)
		{
			// Log it
			log(message, replace);
		}
	}

	/**
	 * Appends a message to the Coords debug console.
	 * @param message the message.
	 */
	public void logCoords(String message) { logCoords(message, false); }
	public void logCoords(String message, boolean replace)
	{
		if(replace == true)
		{
			coordPanel.setText(message);
		}
		else 
		{
			coordPanel.append(message + "\n");
			// Jump to this new line - note; this will error if we want to debug before the scroll box is created
			coordScroll.getVerticalScrollBar().setValue(coordScroll.getVerticalScrollBar().getMaximum());
			coordPanel.select(coordPanel.getText().length(), 0);
		}
	}

	// AA: Altered from Chris Carter's example. This is privately used to print to a log file (or create one if logging was just enabled).
	// It replaced what was running in the general working thread. No co-ordinates are sent to the log file however.
	private void logToFile(String message)
	{
		// Log to file?
		if(saveLog)
		{
			// Create new file with date
			if(logFile == null)
			{
				// Get timestamp and use this as file name
				Date now = new Date();
				String nowFormatted = Utility.formatFileName(now.toString());				
				try{
					// Create log file
					logFile = new FileOutputStream("logs\\"+nowFormatted + ".txt");
					// Convert to print stream
					logStream = new PrintStream( logFile );
					
					// Log the logging creation :P
					infoPanel.append("\nLog File: " + nowFormatted + ".txt\n\n");

					// Write the initial message that is logged after the checkbox is enabled
					logStream.println(message);
					logStream.flush();
				} catch(FileNotFoundException e){
					infoPanel.append("Could not create log file!\n");
					infoPanel.append("Error:" + e.toString() + "\n");
				}
			}
			else
			{
				// Write to existing log file
				logStream.println(message);
				logStream.flush();
			}
		}
	}
	
	public void closeLog()
	{
		// If saving log...
		if(saveLog) {
			// Last logging item needs to note this
			log("Closing log...");

			// Close streams and file
			try {
				logStream.close(); // Also flushes stream
				logFile.close();
			}
			catch (IOException e) {
				System.out.println("Error closing log: " + e.toString());
			}
			// Turn it off
			saveLog = false;
		}
	}
	
	// Turn off car on system exit
	public void windowClosing(WindowEvent windowevent) {
		setVisible(false);
		closeLog();
		if(engine.camera.getCameraOn()){
			camImg.imageGrabber.shutdown();
			camImg.cameraController.shutdown();
		}
		SimCarControl cC = new SimCarControl();
		cC.close(6);
		System.out.println(cC.exitMsg);
		System.exit(0);
	}

	public void windowActivated(WindowEvent e) {}
	public void windowClosed(WindowEvent e) {}
	public void windowDeactivated(WindowEvent e) {}
	public void windowDeiconified(WindowEvent e) {}
	public void windowIconified(WindowEvent e) {}
	public void windowOpened(WindowEvent e) {}
}