/**
 * DrivingAI.java
 * Handles information to decide what action the car should perform using the remote control.
 *
 * How it works is using a loop (as previous projects have done) to interpret changes in the camera
 * and using predictions, to change what pulses are sent to the car. Once turned on, it should attempt
 * a certain action which the control panel allows.
 * 
 * We run the loop on the basis of our main source of information: the camera. This runs at 30 FPS, meaning
 * it is easy to evaluate some timings on how long an command should be sent before we see results.
 * Driving should be quite fluid this way - even if it crashes!
 * 
 * The AI will try and detect walls and will get around the track by trying to reach the nearest waypoint.
 * If crashed (on a wall, or placed the wrong direction) then the AI should attempt a manoeuver to sort itself out.
 * 
 * Planned actions:
 * - Crash recovery
 * - Car navigation around the track
 * - Car racing around the track (combination of above two)
 *
 * Created by: Andrew Armstrong
 * Last Modified: 04/05/08
 *
 */
import parport.ParallelPort;

public class DrivingAI implements Runnable {

	// Navigation AI
	NavigationAI navigationAI;

	// If the AI is on, this is TRUE, otherwise the AI shouldn't run (or will stop when it is running).
	private boolean toggleAI = false;
	// If this is true, it will pause the AI thread. Is still overriden by the stop above.
	private boolean pauseAI = false;
	// If this is TRUE we have setup the AI at least once. We can *never* start without this being true.
	private boolean setupDrivingAIDoneOnce = false;
	// Used to tell if we've not re-setup - if so, it'll warn the user the last values will be used
	private boolean setupDrivingAIDone = false;
	// Only go one point if this is true
	public boolean crashRecoveryOnly = false;
	
	// Threading
	Thread drivingAIThread;
	// For logging
	private CarAIMain parentCarAIMain;
	// Parallel port for actual control	
	private ParallelPort pp = new ParallelPort(0x378);

	// Parallel port control bits
	private static final int leftBit = 0x01; //i.e. data pin 0 of parallel port
	private static final int rightBit = 0x02; //i.e. data pin 1 of parallel port
	private static final int forwardBit = 0x04; //i.e. data pin 2 of parallel port
    private static final int backwardBit = 0x08; //i.e. data pin 3 of parallel port
    //N.B. If 2 cars are to be used, shift above 4 bits to right to access other 4 data pins (for control of second car)

	/*************************
		Constructor
	*************************/

	public DrivingAI(CarAIMain passedCarAIMain, Camera passedCamera) {
		passedCarAIMain.logDetail("[DrivingAI] Constructor: Starting");
		// For logging
		parentCarAIMain = passedCarAIMain;

		// Navigation AI creation
		navigationAI = new NavigationAI(passedCarAIMain, passedCamera);
		passedCarAIMain.logDetail("[DrivingAI] Constructor: Done");
	}

	/*************************
		PUBLIC FUNCTIONS 
	*************************/

	// To run the AI this thread is needed
	public void run() {

		//Determine which Thread is currently running
		Thread thisThread = Thread.currentThread();

		// Do driving actions while thread is active
		while (thisThread == drivingAIThread){
			// Do we want to stop?
			if(toggleAI == false)
			{
				drivingAIThread = null;
				driveStop(0);
				parentCarAIMain.log("[DrivingAI] DrivingAI stopped/paused.");
			}
			else if(pauseAI == true)
			{
				// Paused AI, means we sleep
				try{
					Thread.sleep(1000);
				} catch (InterruptedException e){
					parentCarAIMain.log("[DrivingAI] Error sleeping thread. Error: " + e.toString());
				}
			}
			else
			{		
				// Carry on...
				parentCarAIMain.log("[DrivingAI] DrivingAI loop start. *****************************************");

				// We now know the car location
				// Find out what we want to do now...
				// Constants available:
				// * BACKWARD
				// * FORWARD
				int forwardBackward = navigationAI.FORWARD;
				// Direction:
				// * STRAIGHT
				// * LEFT
				// * RIGHT
				int direction = navigationAI.STRAIGHT;
				
				// Key to how fast we move, since it is the amount of time we hold down "forwards" or "backwards".
				// NOTE: The forward speed is REDICULOUS if held down, this should ALWAYS have a gap between decisions
				int speed = 150;
				// Interval is the total amount of time we sleep() for in milliseconds. interval - speed is the gap between choosing a new action (of course, it takes time to
				// decide what to do as well)
				int interval = 700;

				// Sort updates for navigation
				// * Car location needs to be found
				// * Update what waypoint we should be going to (if it has changed)
				if(navigationAI.updateNavigationAI())
				{
				
					// We need to get the direction and forward/Backwards from the navigation AI
					// Simply because calling a ton of functions means we'll be redoing a lot of things again and again, we 
					// do one simple call to the navigation AI to sort the issue and the capture the results
					if(navigationAI.navigateDecision())
					{
						// Crash recovery check
						if(crashRecoveryOnly)
						{
							// Check if got towards correct facing or are at the point now
							if(navigationAI.crashRecoveryCheck())
							{
								// Report true
								parentCarAIMain.log("[DrivingAI] Navigation AI reports we're facing or are at the point we want to be. Stopping.");
								stopDriving();
							}
						}
						// Crash recovery intervention
						if(toggleAI == true)
						{
							// Get direction needed
							direction = navigationAI.directionChosen;
							forwardBackward = navigationAI.forwardBackwardChosen;

							// check for speed changes
							if(forwardBackward == navigationAI.FORWARD && direction == navigationAI.STRAIGHT)
							{
								// Half speed for directly forwards. This really wizzes!
								speed /= 2;
							}
							else if(forwardBackward == navigationAI.BACKWARD)
							{
								// Add one 50 more for backwards movement, since it usually is a slower momentum
								speed += 50;
							}
							// Drive
							drive(forwardBackward, direction);
						}
					}
					else
					{
						// We can't do anything, we have an error
						parentCarAIMain.log("[DrivingAI] Error! Navigation AI reports no good navigation decision to make! No action.");
					}
				}
				else
				{
					// We can't do anything, we have an error
					parentCarAIMain.log("[DrivingAI] Error! Navigation AI reports failure, cannot find car perhaps. No action.");
				}
				
				// Do the above chosen action for "speed" amount of seconds
				try{
					Thread.sleep(speed);
				} catch (InterruptedException e){
					parentCarAIMain.log("[DrivingAI] Error sleeping thread. Error: " + e.toString());
				}			
				// Stop, but do continue holding down the direction we wanted to "slide" for further
				driveStop(direction);

				// Sleep for interval - speed
				try{
					Thread.sleep(interval - speed);
				} catch (InterruptedException e){
					parentCarAIMain.log("[DrivingAI] Error sleeping thread. Error: " + e.toString());
				}
			}
		}
	}
	
	// Setup the AI - find the waypoints, boundries, and so forth. Generally interfaces with NavigationAI
	public void setupDriving() {
		// Initalise driving AI...
		parentCarAIMain.log("[Driving AI] Initializing (No car should be on the track)...");

		// Setup the navigation AI, and make sure it detected things correctly!
		if(navigationAI.setupRaceNavigationAI())
		{
			// Done, now can start
			setupDrivingAIDone = true;
			// Done it at least once too
			setupDrivingAIDoneOnce = true;
			
			// Toggle buttons
			parentCarAIMain.startAIButton.setEnabled(true);
			parentCarAIMain.incrementStartingPointAIButton.setEnabled(true);

			// Done!
			parentCarAIMain.log("[Driving AI] Finished initializing, ready to start!\nIf you left the car on the track, please redo.");
		}
		else
		{
			// Error!
			parentCarAIMain.log("[Driving AI] Error initializing! Check camera is on.");		
		}
	}

	// Pause / resume driving functionality
	public void pauseDriving() {
		// Have the thread do intervals of sleep
		pauseAI = true;
		// Change button functionality
		parentCarAIMain.pauseResumeAIButton.setText("Resume AI");		
	}
	public void resumeDriving() {
		// Just carry on thread
		pauseAI = false;
		// Change button functionality
		parentCarAIMain.pauseResumeAIButton.setText("Pause AI");
	}
	// Start the AI - turn the loop to the "on" position, and let the next frame work. This does however setup the navigation.
	public void startDriving() {
	
		// Cannot work without setupDrivingAIDoneOnce = true
		if(setupDrivingAIDoneOnce == false)
		{
			parentCarAIMain.log("[Driving AI] Cannot start AI until we have the track setup at least once with no errors.");
			return;
		}

		// Setup navigation (find the car!)
		if(navigationAI.initialRaceNavitationAI())
		{
			// Go go go!
			toggleAI = true;
			
			// Start thread
			drivingAIThread = new Thread(this);
			drivingAIThread.start();

			// Toggle buttons
			parentCarAIMain.pauseResumeAIButton.setEnabled(true);
			resumeDriving(); // Sets the button to "Pause AI" and has the variable set to false
			parentCarAIMain.stopAIButton.setEnabled(true);
			parentCarAIMain.setupAIButton.setEnabled(false);
			parentCarAIMain.incrementStartingPointAIButton.setEnabled(false);
			
			// Debug
			parentCarAIMain.log("[Driving AI] Turned on. #################################");			
			return;
		}
		
		// Debug
		parentCarAIMain.log("[Driving AI] Error: cannot start without car being found.");
	}
	
	// Stop the AI, as noted, will not stop current commands, but stops the loop from operating.
	public void stopDriving() {
		// Stop! Let the loop handle it
		toggleAI = false;
		
		// Toggle buttons
		parentCarAIMain.pauseResumeAIButton.setEnabled(false);
		parentCarAIMain.pauseResumeAIButton.setText("Pause AI");
		parentCarAIMain.stopAIButton.setEnabled(false);
		parentCarAIMain.setupAIButton.setEnabled(true);

		// Debug
		parentCarAIMain.log("[Driving AI] Turned off. #################################");
	}
	
	/*************************
		PRIVATE FUNCTIONS 
	*************************/
	
	// Drive the car, taken from Scott Culcheth's code
    private void drive(int speed) {	drive(speed, navigationAI.STRAIGHT); }
	private void drive(int speed, int dir) {
        // Control bits set direction
        int control = getDriveLeftRight(dir);
        // now add in the necessary bits
        if (speed == navigationAI.FORWARD) {
            control |= forwardBit;
        }
        else if (speed == navigationAI.BACKWARD) {
            control |= backwardBit;
        }
        pp.write(control);
	}

	// Get's the control bit according to how we are driving
	private int getDriveLeftRight(int dir) {
		int returnBits = 0;
		if (dir == navigationAI.STRAIGHT) { // also default if unset
			// Do nothing
			return returnBits;
		}		
		else if (dir == navigationAI.LEFT) {
            returnBits |= leftBit;
        }
        else if (dir == navigationAI.RIGHT) {
            returnBits |= rightBit;
        }
		return returnBits;	
	}
	
	// Stop the car
	private void driveStop(int dir) {
		// Write "stop"
		int control = getDriveLeftRight(dir);
		pp.write(control);
	}
}