/**
 * Engine.java 
 *
 * This will run the core functions of the program. It is created and runs from the entry class CarAIMain, and controls these aspects:
 * - Creates DrivingAI and SimulationAI classes for their functionality.
 * - Program loop for AI, or running a simulation, or whatever is needed - only one at a time however
 *	- It should invoke a classes' "start()" function so that can deal with the threading stuff!
 * - Sorting button presses, and changes in GUI (references back to the initial creation for this)
 * 	- With this, it therefore also operates the camera on/off etc.
 *
 * The main change from the previous project is the removal of direct GUI code from the "SimDraw" class essentially.
 * 
 * For the listeners to work, this extends the Component class although it is not a displayed GUI element.
 *
 * Created by: Andrew Armstrong
 * Last Modified: 04.05.08
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.lang.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.image.*;

public class Engine extends Component implements ActionListener, ChangeListener, ItemListener {

	/***********************
		SETTINGS
	************************/

	// Various settings we need to keep track of independently of the GUI

	/***********************
		OTHER VARIABLES
	************************/
	
	// GUI work needs this
	private CarAIMain parentCarAIMain;
	// A special GUI element we'll use a lot - comes from the above
	private SimDraw parentSimDraw;
	
	// Camera stuff is done by this class
	public Camera camera;
	// Tester for image tracking
	public FollowCarTest followCarTest;
	// Driving AI
	public DrivingAI drivingAI;

	private final static long serialVersionUID = 1L;

	/***********************
		CONSTRUCTOR
	************************/
	
	// Constructor. Initalises all classes that do the work we need.
	public Engine(CarAIMain passedCarAIMain) {
		// GUI work needs this
		parentCarAIMain = passedCarAIMain;
		parentSimDraw = parentCarAIMain.simDraw;
		
		// Create the camera interface class
		camera = new Camera(passedCarAIMain);

		// New followCarTest
		followCarTest = new FollowCarTest(passedCarAIMain, camera);

		// Contains navigation AI etc.
		drivingAI = new DrivingAI(passedCarAIMain, camera);
	}
	
	/***********************
		PUBLIC FUNCTIONS
	************************/
	
	
	/***********************
		PRIVATE FUNCTIONS
	************************/

	/***********************
		GUI
	************************/

	// Get change of clock rate slider
	public void stateChanged(ChangeEvent evt){
	
		// Get the slider which has changed, and it's new value
		JSlider sliderSource = (JSlider) evt.getSource();
		int currentValue = sliderSource.getValue();

		// Speed slider
		if(sliderSource == parentCarAIMain.clockSlider)
		{
			// Value of 0 is invalid, cahange to 1
			if (currentValue == 0){
				currentValue = 1;
				// Set this as the new value
				parentCarAIMain.clockSlider.setValue(1);
				parentCarAIMain.clockSlider.setUI(parentCarAIMain.clockSlider.getUI());
			}
			// Set the GUI to represent this new number so the user knows the value
			Integer cR = new Integer( currentValue );
			parentCarAIMain.clockText.setText(cR.toString());
			// This makes sure we only do an action when we finally finish messing around dragging the clock slider
			if (!sliderSource.getValueIsAdjusting()) {
				// Set the clock rate for the simulation
				parentSimDraw.clockRate = currentValue;

				// Disable driveCar to make sure the simulation doesn't run (or stop it if it is running) if clock rate > 1
				if(currentValue > 1){
					parentCarAIMain.driveCarCB.setSelected(false);
					parentCarAIMain.driveCarCB.setEnabled(false);
					parentSimDraw.driveCar = false;
				}
				else{
					// If we have == 1, we will, if we are not running, enable the simulation driving again.
					if(!parentSimDraw.simRunning()) {
						parentCarAIMain.driveCarCB.setEnabled(true);
						// Set drivable to true
						if(parentSimDraw.simStarted() == false)
						{
							parentCarAIMain.driveCarCB.setSelected(true);
							parentSimDraw.driveCar = true;
						}
					}
				}
			}
		}
		// Red threshold slider
		else if(sliderSource == parentCarAIMain.thresholdRedBarSlider)
		{			
			// Value check - can only be 0 to 256
			if(currentValue < 0 || currentValue > 256)
			{
				if(currentValue < 0) currentValue = 0;
				else currentValue = 256;
				// Set this as the new value
				sliderSource.setValue(currentValue);
				sliderSource.setUI(sliderSource.getUI());
			}
			// Set the value of our setting (and the GUI accompanying text) to currentValue
			parentCarAIMain.thresholdRedBarLabelValue.setText(Integer.toString(currentValue));
			// This makes sure we only do an action when we finally finish messing around dragging the clock slider
			if (!sliderSource.getValueIsAdjusting()) {
				camera.imageThresholder.setRHi(currentValue);
			}
		}
		// Green threshold slider
		else if(sliderSource == parentCarAIMain.thresholdGreenBarSlider)
		{		
			// Value check - can only be 0 to 256
			if(currentValue < 0 || currentValue > 256)
			{
				if(currentValue < 0) currentValue = 0;
				else currentValue = 256;
				// Set this as the new value
				sliderSource.setValue(currentValue);
				sliderSource.setUI(sliderSource.getUI());
			}
			// Set the value of our setting (and the GUI accompanying text) to currentValue
			parentCarAIMain.thresholdGreenBarLabelValue.setText(Integer.toString(currentValue));
			// This makes sure we only do an action when we finally finish messing around dragging the clock slider
			if (!sliderSource.getValueIsAdjusting()) {
				camera.imageThresholder.setGHi(currentValue);
			}
		}
		// Blue threshold slider
		else if(sliderSource == parentCarAIMain.thresholdBlueBarSlider)
		{
			// Value check - can only be 0 to 256
			if(currentValue < 0 || currentValue > 256)
			{
				if(currentValue < 0) currentValue = 0;
				else currentValue = 256;
				// Set this as the new value
				sliderSource.setValue(currentValue);
				sliderSource.setUI(sliderSource.getUI());
			}
			// Set the value of our setting (and the GUI accompanying text) to currentValue
			parentCarAIMain.thresholdBlueBarLabelValue.setText(Integer.toString(currentValue));
			// This makes sure we only do an action when we finally finish messing around dragging the clock slider
			if (!sliderSource.getValueIsAdjusting()) {
				camera.imageThresholder.setBHi(currentValue);
			}	
		}
		else if(sliderSource == parentCarAIMain.backgroundSubtracterSlider)
		{
			// Value check - can only be 0 to 256
			if(currentValue < 0 || currentValue > 256)
			{
				if(currentValue < 0) currentValue = 0;
				else currentValue = 256;
				// Set this as the new value
				sliderSource.setValue(currentValue);
				sliderSource.setUI(sliderSource.getUI());
			}
			// Set the value of our setting (and the GUI accompanying text) to currentValue
			parentCarAIMain.backgroundSubtracterSliderValue.setText(Integer.toString(currentValue));
			// This makes sure we only do an action when we finally finish messing around dragging the clock slider
			if (!sliderSource.getValueIsAdjusting()) {
				camera.backgroundSubtracter.setThreshold(currentValue);
			}	
		}
	}

	// Get change of checkbox
	public void itemStateChanged(ItemEvent evt){
	
		// What checkbox? what is it's status now?
		Object clickedBox = evt.getSource();
		int status = evt.getStateChange();

		// "Drive Car"
		if(clickedBox == parentCarAIMain.driveCarCB){
			parentSimDraw.driveCar = (status == ItemEvent.SELECTED) ? true : false;
		}
		// "Detailed log"
		else if(clickedBox == parentCarAIMain.detInfoCB){
			parentCarAIMain.detInfo = (status == ItemEvent.SELECTED) ? true : false;
			// If at the start, we reset the log with this new detailed information.
			if(parentSimDraw.simStarted() == false){
				parentSimDraw.simLoggingReset();
			}
		}
		// "Save Log"
		else if(clickedBox == parentCarAIMain.saveLogCB){
			if(status == ItemEvent.SELECTED) {
				// Save the log - say we are doing it will initalise the log				
				parentCarAIMain.saveLog = true;
				parentCarAIMain.log("Save Log: Logging started.");
			}
			else {
				// This turns off the logging
				parentCarAIMain.log("Save Log: Logging ending.");
				parentCarAIMain.closeLog();
			}
		}
		// "Stop on crash"
		else if(clickedBox == parentCarAIMain.stopCrashCB){
			if(status == ItemEvent.SELECTED){
				parentSimDraw.stopOnCrash = true;
				parentCarAIMain.log("Stop On Crash Activated");
			}
			else {
				parentSimDraw.stopOnCrash = false;
				parentCarAIMain.log("Stop On Crash Deactivated");
			}
		}
		// "Stop on crash"
		else if(clickedBox == parentCarAIMain.stopCrashCB){
			if(status == ItemEvent.SELECTED){
				parentSimDraw.stopOnCrash = true;
				parentCarAIMain.log("Stop On Crash Activated");
			}
			else {
				parentSimDraw.stopOnCrash = false;
				parentCarAIMain.log("Stop On Crash Deactivated");
			}
		}
		// "Crash Recovery Only"
		else if(clickedBox == parentCarAIMain.crashRecoveryCB){
			if(status == ItemEvent.SELECTED){
				drivingAI.crashRecoveryOnly = true;
				parentCarAIMain.log("Crash Recovery Only Activated");
			}
			else {
				drivingAI.crashRecoveryOnly = false;
				parentCarAIMain.log("Crash Recovery Only Deactivated");
			}
		}
	}

	//Get which button has been clicked and call appropriate method
	public void actionPerformed(ActionEvent event){
		String command = event.getActionCommand();
		
		if (command == "Start Simulation" || command == "Resume Simulation"){
			parentSimDraw.startPlaying();
		}
		else if (command == "Stop Simulation"){
			// Simulation
			parentSimDraw.stopPlaying();
		}
		else if (command == "Reset"){
			// Simulation
			parentSimDraw.resetSim();
		}
		// Normal camera stuff
		else if (command == "Turn On Camera"){
			camera.enableCam();
		}
		else if (command == "Turn Off Camera"){
			camera.disableCam();
		}
		// AA additional for camera image
		else if (command == "Save Live PNG Image") {
			camera.saveLiveCameraImage();
		}
		else if (command == "Save Static PNG Image") {
			camera.saveStaticCameraImage();
		}
		else if (command == "Load Image") {
			camera.loadCameraImage();
		}
		// AI controls
		else if (command == "Setup AI") {
			parentCarAIMain.log("Setup AI button pressed");
			drivingAI.setupDriving();
		}
		else if (command == "Start AI") {
			parentCarAIMain.log("Start AI button pressed");
			drivingAI.startDriving();
		}
		else if (command == "Pause AI") {
			parentCarAIMain.log("Pause AI button pressed");
			drivingAI.pauseDriving();
		}
		else if (command == "Resume AI") {
			parentCarAIMain.log("Resume AI button pressed");
			drivingAI.resumeDriving();
		}
		else if (command == "STOP AI") {
			parentCarAIMain.log("STOP AI button pressed");
			drivingAI.stopDriving();
		}
		else if (command == "Display Past Points") {
			parentCarAIMain.log("Display Past Points AI button pressed");
			drivingAI.navigationAI.displayAllPastPoints();
		}
		else if (command == "Increment Starting point") {
			parentCarAIMain.log("Increment Starting point pressed");
			drivingAI.navigationAI.track.incrementCurrentWaypoint();
		}
		
		// Threshold image button
		else if(command == "Threshold Image") {
			parentCarAIMain.log("Threshold Image...");
			camera.thresholdButton();
		}
		
		// Background subtraction image button
		else if(command == "Save Reference Image") {
			parentCarAIMain.log("Save reference image");
			camera.saveReferenceButton();
		}
		else if(command == "Background Subtraction") {
			parentCarAIMain.log("Background subtraction...");
			camera.backgroundSubtractionButton();
		}
		else if(command == "Frame Difference") {
			parentCarAIMain.log("Frame differencing...");
			camera.frameDifferencing();
		}
		
		// Debug button
		else if(command == "Test Button 1") {
			parentCarAIMain.log("Test Button 1 pressed");
			
			// Work on a live image...
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
			
			// Draw at waypoint
			Coordinate waypoint = new Coordinate(100, 100);
			Coordinate newCoord;
			int newBlockValue;
			Color color;
			double waypointAngle, waypointAngle2, differenceAngle, differenceAngle2, differenceAngle3, differenceAngle4;
			// Draw
			camera.drawBlock(waypoint, liveImage, Color.RED);

			// Report some numbers
			for(int i = 0; i <= 360; i += 45)
			{
				newCoord = Utility.getProjectedPoint(waypoint, i, 20, 0, 320, 0, 240);
				// Draw depending on angle
				color = Color.BLUE;
				switch(i)
				{
					case 0:		color = Color.BLUE; break;
					case 45:	color = Color.CYAN; break;
					case 90:	color = Color.YELLOW; break;
					case 135:	color = Color.WHITE; break;
					case 180:	color = Color.GREEN; break;
					case 225:	color = Color.LIGHT_GRAY; break;
					case 270:	color = Color.MAGENTA; break;
					case 315:	color = Color.ORANGE; break;
					case 360:	color = Color.PINK; break;				
				}

				camera.drawBlock(newCoord, liveImage, color);

				// i = waypoint angle, j = car angle
				for(int j = 0; j <= 360; j += 45)
				{
					// Report angle
					waypointAngle = Utility.calcAngle(waypoint, newCoord);
					// Difference of angles...we know our directions but not where we want to turn!
					differenceAngle = j - waypointAngle;
					differenceAngle2 = waypointAngle - j;

					waypointAngle2 = Utility.calcAngle(newCoord, waypoint);
					differenceAngle3 = j - waypointAngle2;
					differenceAngle4 = waypointAngle2 - j;
					
					// Report
					parentCarAIMain.log("[Debug] newCoord (" + newCoord.x +"," + newCoord.y + ") at angle from waypoint [" + waypointAngle + "] with facing [" + j + "] differenceAngle ["+differenceAngle+"] differenceAngle2 [" + differenceAngle2 + "] waypointAngle2 [" + waypointAngle2 + "] differenceAngle3 ["+ differenceAngle3 + "] differenceAngle4 [" + differenceAngle4 +"]");
				}
			}

			// Display
			parentCarAIMain.staticCamImg.image.setImage(liveImage);
			
			//drivingAI.navigationAI.displayAllPastPoints();
			
			//long start, stop, elapsed;
	        //double freq;
			
			//parentCarAIMain.log("Finding car");
			//start = System.currentTimeMillis();
			
			
			//drivingAI.navigationAI.car.getCarPosition();
			
			
			//stop = System.currentTimeMillis();
			//elapsed = stop - start;
			//parentCarAIMain.log("Time taken:" + "\t" + elapsed);
			//parentCarAIMain.log("Finding done");
/*		
			parentCarAIMain.log("Live Image sizes: W:" + Integer.toString(parentCarAIMain.camImg.liveImage.getImageWidth()) + " H: " + Integer.toString(parentCarAIMain.camImg.liveImage.getImageHeight()));
			BufferedImage test = parentCarAIMain.camImg.getLiveImage();
			parentCarAIMain.log("Test size: W:" + Integer.toString(test.getWidth()) + " H: " + Integer.toString(test.getHeight()));
*/
		}
		else if(command == "Test Button 2") {
			parentCarAIMain.log("Test Button 2 pressed");
			
			
			// Work on a live image...
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
			
			camera.drawLine(0, 0, 50, 40, liveImage, Color.RED);

			// Display
			parentCarAIMain.staticCamImg.image.setImage(liveImage);
			
		    //long start, stop, elapsed;
	        //double freq;
			//start = System.currentTimeMillis();
			
			
			//drivingAI.navigationAI.car.findCarFromPastCoordinates();
			
			
			//stop = System.currentTimeMillis();
			//elapsed = stop - start;
			//parentCarAIMain.log("Time taken:" + "\t" + elapsed);
			//parentCarAIMain.log("Finding done");
		}
		else if(command == "Test Button 3") {
			parentCarAIMain.log("Test Button 3 pressed");
			
			// Work on a live image...
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
			
			// Randomise!
			///*
			Coordinate coord1 = new Coordinate(Utility.randomInt(10, 100), Utility.randomInt(10, 100));
			Coordinate coord2 = new Coordinate(Utility.randomInt(10, 100), Utility.randomInt(10, 100));
			Coordinate coord3 = Utility.getProjectedPoint(coord1, coord2, Utility.randomInt(5, 100), 0, 320, 0, 240);
			//*/
			/*
			Coordinate coord1 = new Coordinate(30, 30);
			Coordinate coord2 = new Coordinate(20, 30);
			Coordinate coord3 = Utility.getProjectedPoint(coord1, coord2, 20, 0, 320, 0, 240);
			//*/
			
			// End up with an angle 0-259.999
			double angle = Utility.calcAngle(coord1.x, coord1.y, coord2.x, coord2.y);
			
			// Coordinates
			camera.drawBlock(coord1.x, coord1.y, liveImage, Color.RED);
			camera.drawBlock(coord2.x, coord2.y, liveImage, Color.YELLOW);
			camera.drawBlock(coord3.x, coord3.y, liveImage, Color.GREEN);

			parentCarAIMain.log("Points done: 1: (" + coord1.x + "," + coord1.y + ") 2 (" + coord2.x + "," + coord2.y + ") 3 (" + coord3.x + "," + coord3.y + ") angle: " + angle);
			parentCarAIMain.log("Distance between points. 1 and 2 [" + Utility.getDistanceBetween(coord1, coord2) + "] 2 and 3 [" + Utility.getDistanceBetween(coord2, coord3) + "] 1 and 3 ["+ Utility.getDistanceBetween(coord1, coord3) + "]");
			
			// Display
			parentCarAIMain.staticCamImg.image.setImage(liveImage);
		}
		else if(command == "Test Button 4") {
			parentCarAIMain.log("Test Button 4 pressed");
			
			// Work on a live image...
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
			
			// Randomise!
			///*
			Coordinate coord1 = new Coordinate(Utility.randomInt(10, 100), Utility.randomInt(10, 100));
			Coordinate coord2 = new Coordinate(Utility.randomInt(10, 100), Utility.randomInt(10, 100));
			
			// Draw all points between an arbitary end point 3
			int distance = Utility.randomInt(5, 100);
			Coordinate coord3 = Utility.getProjectedPoint(coord1, coord2, distance, 0, 320, 0, 240);

			// Draw line between 1 and 2
			// End up with an angle 0-259.999
			double angle = Utility.calcAngle(coord1.x, coord1.y, coord2.x, coord2.y);
			// Sort cos and sin			
			double xMultiply = Math.cos(Math.toRadians(angle));
			double yMultiply = Math.sin(Math.toRadians(angle));
			int xDraw, yDraw;
			for(int i = 1; i <= distance; i++)
			{			
				// Calculate this point to draw
				xDraw = Utility.getBoundriedInt(coord2.x + (int)(i * xMultiply), 0, 320);
				yDraw = Utility.getBoundriedInt(coord2.y + (int)(i * yMultiply), 0, 240);
				
				// Log
				parentCarAIMain.log("Line: X,Y (" + xDraw + "," + yDraw + ")");

				// Draw it
				liveImage.setRGB(xDraw, yDraw, Color.BLUE.getRGB());
			}
			
			//*/
			/*
			Coordinate coord1 = new Coordinate(30, 30);
			Coordinate coord2 = new Coordinate(20, 30);
			Coordinate coord3 = Utility.getProjectedPoint(coord1, coord2, 20, 0, 320, 0, 240);
			//*/
			
			// Coordinates
			camera.drawBlock(coord1.x, coord1.y, liveImage, Color.RED);
			camera.drawBlock(coord2.x, coord2.y, liveImage, Color.YELLOW);
			camera.drawBlock(coord3.x, coord3.y, liveImage, Color.GREEN);

			parentCarAIMain.log("Points done: 1: (" + coord1.x + "," + coord1.y + ") 2 (" + coord2.x + "," + coord2.y + ") 3 (" + coord3.x + "," + coord3.y + ") angle: " + angle);
			parentCarAIMain.log("Distance between points. 1 and 2 [" + Utility.getDistanceBetween(coord1, coord2) + "] 2 and 3 [" + Utility.getDistanceBetween(coord2, coord3) + "] 1 and 3 ["+ Utility.getDistanceBetween(coord1, coord3) + "]");
			
			// Display
			parentCarAIMain.staticCamImg.image.setImage(liveImage);
			
//			camera.Test_CheckTopLeftColour();
		}
		else if(command == "Test Following Start") {
			parentCarAIMain.log("Test Following Start button pressed");
			followCarTest.followCarTestStart();
		}
		else if(command == "Test Following End") {
			parentCarAIMain.log("Test Following End  button pressed");
			followCarTest.followCarTestEnd();
		}
	}
}