/**
 * This will do thresholding on an image. Anything below a certain point is "marked" - anything above is empty.
 *
 * The main use of this is to get the black track to show up as full "Green" and everything else as black.
 *
 * This makes it so that anything that isn't green is "impassible" and an obstacle. This works well too for detecting the
 * car, especially when combined with frame differencing.
 *
 * Modified from original to not need stored "live" and "threshold" images, which are brought in when needed instead.
 * Sliders modify the values - noting the "Blue" is more important then Red or Green.
 * 
 * Original Code: David Bull - http://www.uk-dave.com/projects/java/colourtracker.php
 * Modified by: Scott Culcheth
 * Version: 0.04
 *
 * Further Modification: Andrew Armstrong
 * Last Modified: 04.05.08
 */
import java.util.*;
import java.awt.*;
import java.awt.image.*;

public class ImageThresholder
{
	////////////////////////////////////////////////////
	// CONSTANTS AND CLASS VARIABLES
	////////////////////////////////////////////////////

	// Colours to change for things under and over the threshold.
	public static final int EMPTY_REGION = 0x000000;// Black, when any red/green/blue are OVER the threshold values
	public static final int FULL_REGION  = 0x00FF00;// Green, when ALL red/green/blue are BELOW the threshold values

	// For logging
	private CarAIMain parentCarAIMain;

	// Default values for the GUI
	// Was 190, 190, 210 (seems to work only at all in very very high light)
	// Lower light: 140, 140, 115 works well.
	private int rHi = 140;
	private int gHi = 140;
	private int bHi = 140;

	////////////////////////////////////////////////////
	// CONSTRUCTOR
	////////////////////////////////////////////////////

	public ImageThresholder(CarAIMain passedCarAIMain)
	{
		parentCarAIMain = passedCarAIMain;
	}

	// Process and return the thresholded image
	// No input uses a live image
	public BufferedImage applyThresholdToImage()
	{
		return applyThresholdToImage(parentCarAIMain.camImg.getLiveImage());
	}

	// This uses inputImage as a base to threshold
	public BufferedImage applyThresholdToImage(BufferedImage inputImage) { return applyThresholdToImage(inputImage, 0, inputImage.getWidth(), 0, inputImage.getHeight());  }
	public BufferedImage applyThresholdToImage(BufferedImage inputImage, int minX, int maxX, int minY, int maxY)
	{
		//parentCarAIMain.log("applyThresholdToImage. Image: " + inputImage.toString() + ". minX [" + minX + "] maxX [" + maxX + "] minY [" + minY + "] maxY [" + maxY + "]");
	
		// Return image
		BufferedImage thresholdImage = new BufferedImage(inputImage.getWidth(), inputImage.getHeight(), BufferedImage.TYPE_INT_RGB);

		// Variables for loop
		int rgb=0, r=0, g=0, b=0;

		// Loop every pixel
		for (int x=minX; x<maxX; x++)
		{
			for (int y=minY; y<maxY; y++)
			{
				// Check RGB values separately.
				rgb = inputImage.getRGB(x,y);
				r = (rgb & 0xFF0000) >> 16;
				g = (rgb & 0x00FF00) >> 8;
				b = (rgb & 0x0000FF);
				
				// Need to have red, green and blue under the threshold values to fillt he region to full, green
				if (r<=rHi && g<=gHi && b<=bHi)
					thresholdImage.setRGB(x,y,FULL_REGION);
				/*else
					// Any pixel which has any of the 3 values over the threshold levels is marked as empty, black.
					// Not needed for a new BufferedImage since it defaults to black.
					thresholdImage.setRGB(x,y,EMPTY_REGION);
				*/
			}
		}
		return thresholdImage;
	}

	// This will threshold "inputImage" directly.
	public void applyThresholdDirectlyToImage(BufferedImage inputImage) { applyThresholdDirectlyToImage(inputImage, 0, inputImage.getWidth(), 0, inputImage.getHeight());  };
	public void applyThresholdDirectlyToImage(BufferedImage inputImage, int minX, int maxX, int minY, int maxY)
	{
		//parentCarAIMain.log("applyThresholdDirectlyToImage. Image: " + inputImage.toString() + ". minX [" + minX + "] maxX [" + maxX + "] minY [" + minY + "] maxY [" + maxY + "]");
	
		// Variables for loop
		int rgb=0, r=0, g=0, b=0;
		// Loop every pixel
		for (int x=minX; x<maxX; x++)
		{
			for (int y=minY; y<maxY; y++)
			{
				// Check RGB values separately.
				rgb = inputImage.getRGB(x,y);
				r = (rgb & 0xFF0000) >> 16;
				g = (rgb & 0x00FF00) >> 8;
				b = (rgb & 0x0000FF);

				// Need to have red, green and blue under the threshold values to fillt he region to full, green
				if (r<=rHi && g<=gHi && b<=bHi)
					inputImage.setRGB(x,y,FULL_REGION);
				else
					// Any pixel which has any of the 3 values over the threshold levels is marked as empty, black
					// Required since inputImage might be some old "live" image
					inputImage.setRGB(x,y,EMPTY_REGION);
			}
		}
	}

	// Get/Set values for the slider values
	public int getRHi() { return rHi; }
	public int getGHi() { return gHi; }
	public int getBHi() { return bHi; }

	public void setRHi(int r) { if (r>=0 && r<=256) rHi=r; }
	public void setGHi(int g) { if (g>=0 && g<=256) gHi=g; }
	public void setBHi(int b) { if (b>=0 && b<=256) bHi=b; }
}
