/**
 * JImagePanel.java
 *
 * Panel object for holding CameraImage object within the interface defined by Simulation.java
 *
 * Original Code: David Bull - http://www.uk-dave.com/projects/java/colourtracker.php
 * Modified by: Scott Culcheth
 * Further modified by: Chris Carter
 * Further modified by: Andrew Armstrong
 *  
 * Last Modified: 04.05.08
 *
 */

import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
// Image loading stuff
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import javax.swing.event.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.*;

public class JImagePanel extends JPanel implements SwingConstants, Scrollable
{
	////////////////////////////////////////////////////
	// CONSTANTS AND CLASS VARIABLES
	////////////////////////////////////////////////////

	private BufferedImage image;
	private float scaleFactor=1.0f;

	private final static long serialVersionUID = 1L;

	public boolean camOn = false;
	ImageIcon camOffPic = new ImageIcon("camOff.png");
	private JLabel camOffLbl = new JLabel(camOffPic);
	BufferedImage camOffImage;


	////////////////////////////////////////////////////
	// CONSTRUCTOR
	////////////////////////////////////////////////////

	/**
	 * Creates a image panel.
	 */
	public JImagePanel()
	{
		super(true);
		// Image will default to "camOff.png"
		File file = new File("camOff.png");
		try {
			// Read from a PNG file
			camOffImage = ImageIO.read(file);
		}
		catch (IOException e) {
			// Oops, logging not in this class (ever?)
			//parentCarAIMain.log("Error loading camOff.png file. Error: " + e.toString());
			System.out.println("Error loading camOff.png file in JPanelImage. Error: " + e.toString());
		}
		image = camOffImage;
		setBackground(Color.gray);
	}

	////////////////////////////////////////////////////
	// METHODS
	////////////////////////////////////////////////////

	/**
	 * Paint method to draw the image.
	 * @param g
	 */
	public void paint(Graphics g)
	{
		super.paint(g);
		if(camOn){
			if (image!=null) // redundant?
			{
				//this.remove(camOffLbl);
				g.drawImage(image, 0, 0, (int)(image.getWidth()*scaleFactor),  (int)(image.getHeight()*scaleFactor), null);
			}
		}
		else {
			this.add(camOffLbl);
		}
	}

	// Remove camera image when camera is disabled
	public void noImage(){
		image = camOffImage;
		repaint();
	}

	/**
	 * Changes the scale factor for the image.
	 * When the scale factor is changed the image is automatically repainted.
	 * @param scaleFactor the new scale factor.
	 */
	public void setScaleFactor(float scaleFactor)
	{
		this.scaleFactor = scaleFactor;
		Dimension dim = new Dimension((int)(image.getWidth()*scaleFactor),  (int)(image.getHeight()*scaleFactor));
		setMinimumSize(dim);
		setPreferredSize(dim);
		setMaximumSize(dim);
		repaint();
		revalidate();
	}


	/**
	 * Returns the scale factor tor the image.
	 * @return the scale factor tor the image.
	 */
	public float getScaleFactor()
	{
		return scaleFactor;
	}


	/**
	 * Returns a copy of the current image.
	 * @return a copy of the current image.
	 */
	public BufferedImage getImage()
	{
		// AA: Was uncommented now not! (fixed; someone had = instead of ==, sigh)
		if(!camOn)
		{
			//Use camOffImage
			BufferedImage newImage = new BufferedImage(camOffImage.getWidth(), camOffImage.getHeight(), BufferedImage.TYPE_INT_RGB);
			for (int y=0; y<camOffImage.getHeight(); y++)	// scan rows
				for (int x=0; x<camOffImage.getWidth(); x++)		// scan pixels in current row
					newImage.setRGB(x,y,camOffImage.getRGB(x,y));	// copy pixel
			return newImage;
		}
		else
		{
			synchronized (image)
			{
				BufferedImage newImage = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
				for (int y=0; y<image.getHeight(); y++)	// scan rows
					for (int x=0; x<image.getWidth(); x++)		// scan pixels in current row
						newImage.setRGB(x,y,image.getRGB(x,y));	// copy pixel
				return newImage;
			}
		}
	}

	/**
	 * Loads a new image.
	 * @param image the new image.
	 */
	public void setImage(BufferedImage image)
	{
		if(camOn){
			synchronized (image)
			{
				this.image = image;
				if (image==null) scaleFactor=1f;
				Dimension dim = new Dimension((int)(getImageWidth()*scaleFactor),  (int)(getImageHeight()*scaleFactor));
				setMinimumSize(dim);
				setPreferredSize(dim);
				setMaximumSize(dim);
				repaint();
				revalidate();
			}
		} else {
			this.image = null;
			repaint();
		}
	}


	/**
	 * Returns the width of the current image, or 0 if no image loaded.
	 * @return the width of the current image, or 0 if no image loaded.
	 */
	public int getImageWidth()
	{
		if (image!=null)
			return image.getWidth();
		else
			return 0;
	}


	/**
	 * Returns the height of the current image, or 0 if no image loaded.
	 * @return the height of the current image, or 0 if no image loaded.
	 */
	public int getImageHeight()
	{
		if (image!=null)
			return image.getHeight();
		else
			return 0;
	}


	////////////////////////////////////////////////////
	// SCROLLABLE IMPLEMENTATION
	////////////////////////////////////////////////////

	/**
	 * Returns the dimensions of the current image.
	 * @return the dimensions of the current image.
	 */
	public Dimension getPreferredScrollableViewportSize()
	{
		return new Dimension((int)(getImageWidth()*scaleFactor), (int)(getImageHeight()*scaleFactor));
	}


	/**
	 * Causes the panel to scroll in 1/8 image sized blocks.
	 * @return 1/8 image size;
	 */
	public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction)
	{
		if (orientation==HORIZONTAL)
			return (int)(getImageWidth()*scaleFactor)/8;
		else
			return (int)(getImageHeight()*scaleFactor)/8;
	}


	/**
	 * Returns <code>false</code>.
	 * @return <code>false</code>.
	 */
	public boolean getScrollableTracksViewportHeight()
	{
		 return false;
	}


	/**
	 * Returns <code>false</code>.
	 * @return <code>false</code>.
	 */
	public boolean getScrollableTracksViewportWidth()
	{
		return false;
	}


	/**
	 * Causes the panel to scroll in 1/16 image sized blocks.
	 * @return 1/16 image size;
	 */
	public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction)
	{
		if (orientation==HORIZONTAL)
			return (int)(getImageWidth()*scaleFactor)/16;
		else
			return (int)(getImageHeight()*scaleFactor)/16;
	}
}