/**
 * NavigationAI.java
 * 
 * Information for the AI to navigate - this deals with where the car is, where it wants to be,
 * and where it should end up.
 *
 * Navigation is based around nodes set on the track (which are visible in the debug camera).
 * These nodes are inialised by this class once the walls are setup.
 * They are then used to check which way the car should be going, once the class knows the cars
 * current facing. Walls are of course checked - we might have crashed or may be facing the wrong way!!!
 *
 * Facing is also kept in recent memory. Only the initial user-initiated "sort ourselves out to get going"
 * otherwise tests the car's facing.
 *
 * Waypoints are based on "getting close" - once we are a reasonable distance away from the next
 * point we move on to getting to the one after it. This will need testing to get right.
 *
 * Created by: Andrew Armstrong
 * Last Modified: 04/05/08
 *
 */
import java.util.*;
import java.awt.*;
import java.awt.image.*;

public class NavigationAI {

	// BufferedImage stores a threshold image of the track
	private BufferedImage trackThresholdImage;

	// We need to keep track of the track using this
	Track track;
	
	// Need to track the car!
	Car car;

	// Logging and ease of use
	private CarAIMain parentCarAIMain;
	private Camera camera;

	// Memory:	
	// Our "last action" ID. 0 == nothing.
	// Unused
	int lastAction = 0;

	// Completed laps recorded
	private int lapsCompleted = 0;
	// Checked for crash AI
	boolean waypointChanged = false;
	boolean usedCorrectFacing = false;
	
	// An array of past locations from each update. Not stored on the track or car classes!
	private Coordinate[] pastLocationsArray; // Java doesn't like variable arrays, so we cap at last 100 points
	private int pastLocationsArrayIndex = 0;
	private boolean pastLocationsArrayOver100 = false;
	
	// These are retrieved by drivingAI, public variables for now.
	public int directionChosen = 0;
	public int forwardBackwardChosen = 0;
	
	// Used to determine car control to use
    static final int BACKWARD = 0;
    static final int FORWARD = 1;
	// Left/right/straight
	static final int STRAIGHT = 0;
    static final int RIGHT = 1;
    static final int LEFT = 2;
	
	/*************************
		Constructor
	*************************/

	public NavigationAI(CarAIMain passedCarAIMain, Camera passedCamera) {
		passedCarAIMain.logDetail("[NavigationAI] Constructor: Starting");

		// For logging/GUI:
		parentCarAIMain = passedCarAIMain;
		this.camera = passedCamera;
		// New track instance
		track = new Track(passedCarAIMain, passedCamera);
		// New car instance
		car = new Car(passedCarAIMain, passedCamera);
		
		parentCarAIMain.logDetail("[NavigationAI] Constructor: Done");
	}

	/*************************
		PUBLIC FUNCTIONS 
	*************************/

	// Setup the navigation and track tracking via. thresholding
	// Returns TRUE if no errors
	public boolean setupRaceNavigationAI() {

		parentCarAIMain.log("[Navigation AI] Setting up Navigation AI");
		
		// Camera needs to be on
		if(camera.getCameraOn())
		{			
			// Work on a thresholded image we grab and save now
			trackThresholdImage = camera.imageThresholder.applyThresholdToImage();	

			// This is a debug version (might/might not be displayed)
			BufferedImage debugImage = camera.copyImage(trackThresholdImage); 

			// Get the track setup
			if(setupAllTrack(trackThresholdImage, debugImage))
			{			
				// Get the waypoints setup
				if(track.setupWaypoints(debugImage))
				{
					// Done!
					parentCarAIMain.log("[Navigation AI] Setting up Navigation AI: Complete");

					// Display debug image
					parentCarAIMain.staticCamImg.image.setImage(debugImage);

					return true;
				}
			}
			// Display debug image
			parentCarAIMain.staticCamImg.image.setImage(debugImage);
		}
		else
		{
			// Failure
			parentCarAIMain.log("[Navigation AI] setupRaceNavigationAI: Camera not on.");
		}
		// Failure
		parentCarAIMain.log("[Navigation AI] setupRaceNavigationAI: Failure.");
		
		// Error at some point
		return false;
	}
	
	// Button: find track
	public boolean setupAllTrack(BufferedImage thresholdImage, BufferedImage debugImage)
	{
		// Log
		parentCarAIMain.log("[NavigationAI] setupAllTrack");	
	
		// Pass into these things in order
		
		// Main track area
		if(track.findTrackArea(thresholdImage, debugImage, car))
		{
			// Chicane
			if(track.findChicane(thresholdImage, debugImage))
			{		
				// Barrier
				if(track.findBarrier(thresholdImage, debugImage))
				{
					// Walls - smaller area then the track (and needs chicane to work)
					if(track.findWalls(thresholdImage, debugImage))
					{
						// Success!
						parentCarAIMain.log("[Navigation AI] test_ButtonFindTrack: Success. Testing points");

						// Return success
						return true;
					}
				}
			}
		}
		// Return failure
		return false;
	}
	
	// This needs to be called before we start - the current waypoint is likely not the one we actually want.
	public boolean initialRaceNavitationAI() {
		parentCarAIMain.log("[Navigation AI] initialRaceNavitationAI. Starting up Navigation AI (finding car, etc)");

		// Camera on?
		if(camera.getCameraOn())
		{
			// Images needed for this update
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
			// Do thresholding first on live image (we need a new buffered image for this) (only on needed area)
			BufferedImage image = camera.imageThresholder.applyThresholdToImage(liveImage, track.trackMinX, track.trackMaxX, track.trackMinY, track.trackMaxY);
			// Frame Differencing
			FrameDifferencing.frameDifferenceDirectly(trackThresholdImage, image, track.trackMinX, track.trackMaxX, track.trackMinY, track.trackMaxY);

			// We need to find the car
			if(car.getCarPosition(liveImage, image))
			{
				// Got the car location, store as initial location (and therefore delete array by creating a new coordiate array for it)
				pastLocationsArrayIndex = 0;
				pastLocationsArray = new Coordinate[100]; // Java doesn't like variable arrays, so we cap at last 100 points
				pastLocationsArray[pastLocationsArrayIndex] = new Coordinate(car.lastCarCoords.x, car.lastCarCoords.y);
				pastLocationsArrayOver100 = false;
				// Disable button
				parentCarAIMain.displayPastPointsAIButton.setEnabled(false);
				// Laps tracking
				lapsCompleted = 0;
				// Crash recovery needs this
				waypointChanged = false;
				usedCorrectFacing = false;

				parentCarAIMain.log("[Navigation AI] initialRaceNavitationAI: Success! Car found.");
				return true;
			}
		}
		// Debug
		parentCarAIMain.log("[Navigation AI] initialRaceNavitationAI: Error!");
		
		// Error
		return false;
	}

	// This is called from the loop to work out if:
	// - We're on the track at all, or if we can even find the car (error if we can't find the car!)
	// - Our new facing and location (updating the "Car.java" instance)
	// - The next waypoint's location
	// This information can then be used by the driving AI to work out what to do (depends on what we want to do!)
	public boolean updateNavigationAI() {
	
		// Naviation AI...update all variables

		// Images needed for this update
		BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();
		// Do thresholding first on live image (we need a new buffered image for this) (only on needed area)
		BufferedImage image = camera.imageThresholder.applyThresholdToImage(liveImage, track.trackMinX, track.trackMaxX, track.trackMinY, track.trackMaxY);
		// Frame Differencing
		FrameDifferencing.frameDifferenceDirectly(trackThresholdImage, image, track.trackMinX, track.trackMaxX, track.trackMinY, track.trackMaxY);

		// Car location
		if(car.findCarAroundPastCoordinates(liveImage, image))
		{
			// Check for lap finishes with this
			int oldWaypoint = track.getCurrentWaypoint();
		
			// Worked, now we need to check the waypoint status
			track.checkWaypoint(car.lastCarCoords);
			
			// Paint waypoint on image
			camera.drawBlock(track.getCurrentWaypointLocation(), image, Color.PINK);

			// Put up debug image
			parentCarAIMain.staticCamImg.image.setImage(image);
			
			// Check for lap finishes now
			if(track.getCurrentWaypoint() < oldWaypoint)
			{
				// New lap, report
				lapsCompleted++;
				parentCarAIMain.log("[Navigation AI] updateNavigationAI() - New lap started. Completed laps total: " + lapsCompleted);
			}
			// Crash recovery work
			if(oldWaypoint != track.getCurrentWaypoint())
			{
				// Changed
				waypointChanged = true;
			}
			else
			{
				waypointChanged = false;
			}
			
			// Report success
			return true;
		}
		// Failure at some point
		return false;
	}
	
	// Checks if the last point has been reached, or if we are directly facing the point
	public boolean crashRecoveryCheck() {
		// waypointChanged check
		if(waypointChanged)
		{
			return true;
		}
		// Check facing
		if(usedCorrectFacing)
		{
			return true;
		}
		return false;
	}
	
	// We need to decide what to do when navigating. This takes the decision and passes the variables to directionChosen and forwardBackwardChosen.
	public boolean navigateDecision() {
		// We have the car's location and facing
		
		// Add the new car loaction into our waypoint array
		addToPastLocationsArray(car.lastCarCoords);
		
		// If we get stuck we'll stop after 10 seconds of being stuck - not much distance moved!
		double pastDistance = distanceTravelledRecently();

		// Debug: Report distance
		parentCarAIMain.logDetail("[Navigation AI] navigateDecision(). pastDistance [" + pastDistance + "] pastLocationsArrayIndex [" + pastLocationsArrayIndex + "]");

		if(pastDistance > 0 && pastDistance < 5)
		{
			// Error encounted, don't do anything...
			parentCarAIMain.logDetail("[Navigation AI] navigateDecision(). Error! No progress made! pastDistance is [" + pastDistance + "] pastLocationsArrayIndex [" + pastLocationsArrayIndex + "]");
			return false;
		}
		
		// Check where it is in relation to walls.
		// We use angles based off the car facing to the right.
		// 	BL (210)	Left (270)	FL (330)
		// 	Back (180)			Forward (0)
		//	BR (150)	Right (90)	FR (30)
		// Add the calculated car facing to get these angles (doesn't matter if it goes over 360 of course).
		// Maximum returned value is 150 pixels, which is all we need
		// Note on car sizes: we are around 
		// Additional; also need back/forward left/right to make sure we can turn in those directions - we use angles of 30 degrees.
		int wallFrontDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 0);
		int wallFrontRightDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 30);
		int wallRightDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 90);
		int wallBackRightDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 150);
		int wallBackDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 180);
		int wallBackLeftDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 210);
		int wallLeftDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 270);
		int wallFrontLeftDistance = (int)getDistanceToNearestObstacle(car.lastCarCoords, car.lastCarFacing + 330);
		
		// Need angle to the next waypoint
		double waypointAngle = Utility.calcAngle(car.lastCarCoords, track.getCurrentWaypointLocation());

		// We now need to decide the relative limits of left/right so we can decide which way to turn.
		// waypointAngle to waypointAngle + 180 == We are facing to the right, so need to turn left
		// waypointAngle to waypointAngle - 180 == We are facing to the left, so need to turn right
		// We also have the fact we can overlap with "0" or "360", it will loop around (so we we have "right" as from "190 - 370" it will really take 190 - 360, then 0-10 too.
		double rightAngleLimit = waypointAngle + 180;
		double leftAngleLimit = waypointAngle - 180;

		// Check these distances, they determine what we want to do...
		parentCarAIMain.log("[Navigation AI] navigateDecision(). Distances to obstacles. F [" + wallFrontDistance + "] FR [" + wallFrontRightDistance + "] R [" + wallRightDistance + "] BR [" + wallBackRightDistance + "] B [" + wallBackDistance + "] BL [" + wallBackLeftDistance + "] L [" + wallLeftDistance + "] FL [" + wallFrontLeftDistance + "]. Car facing [" + car.lastCarFacing + "] Current Waypoint [" + track.getCurrentWaypoint() + "] and waypointAngle [" + waypointAngle + "]");
		
		// Error checking
		if(wallFrontDistance <= 30 && wallBackDistance <= 30)
		{
			parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front and behind us, cannot move!");
		}

		// The usedCorrectFacing flag is used for crash recovery - if we do move forwards, we must be facing the point (more or less)
		usedCorrectFacing = false;
		
		// ********************************
		// FACING A WALL
		// ********************************
		
		// Facing a wall...
		// This is the biggest barrier to progress, of course!
		if(wallFrontDistance <= 30)
		{
			parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front of us");
			
			// We want to reverse, but what direction?
			
			// If the waypoint is entirely the wrong direction, we always turn to the RIGHT since the track is based on going around almost entirely right
			if(getAreOppositeFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
			{
				// Reverse LEFT, since this will make us able to turn RIGHT more
				forwardBackwardChosen = BACKWARD;
				directionChosen = LEFT;
				parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front of us. We are facing wrong way, so reversing left.");
				return true;
			}
			// If we are actually facing right at the target, we need to turn away from the given wall
			else if(getAreCorrectFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
			{
				parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front of us. We are facing waypoint, avoiding the wall by reversing.");
				// Which side has more space? Move backwards and opposite to that direction
				forwardBackwardChosen = BACKWARD;
				if(wallRightDistance > wallLeftDistance)
				{
					directionChosen = LEFT;
				}
				else
				{
					directionChosen = RIGHT;
				}				
				return true;
			}
			else if(getAreRightFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
			{
				parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front of us. We are on right side of waypoint. Turning back/right to turn left later.");
				// For now, if we are on the right side we should just reverse right (since this gets us more "left")
				forwardBackwardChosen = BACKWARD;
				directionChosen = RIGHT;
				return true;
			}
			else
			{
				parentCarAIMain.log("[Navigation AI] navigateDecision() Wall in front of us. We are on left side of waypoint. Turning back/left to turn right later.");
				// Opposite if we are on the left side, we want to turn left now to get into a position where in the future we can turn right
				forwardBackwardChosen = BACKWARD;
				directionChosen = LEFT;
				return true;
			}
		}

		// ********************************
		// STANDARD NAVIGATION
		// ********************************

		// This occurs if there are no given direct obstacles to our progress

		// If the waypoint is entirely the wrong direction, we always turn to the RIGHT since the track is based on going around almost entirely right
		if(getAreOppositeFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
		{
			// It is on the right of us
			if(getAreRightFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
			{
				if(wallBackDistance > 30)
				{
					// Reverse opposite location
					forwardBackwardChosen = BACKWARD;
					directionChosen = RIGHT;
					parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are facing wrong way, it's on the left side, reverse back right.");
					return true;
				}
				else
				{
					// Move forwards
					forwardBackwardChosen = FORWARD;
					directionChosen = LEFT;
					parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are facing wrong way, it's on right side, forward left.");
					return true;
				}
			}
			else
			{
				if(wallBackDistance > 30)
				{
					// Reverse opposite location
					forwardBackwardChosen = BACKWARD;
					directionChosen = LEFT;
					parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are facing wrong way, it's on left side, reverse back left.");
					return true;
				}
				else
				{
					// Move forwards
					forwardBackwardChosen = FORWARD;
					directionChosen = RIGHT;
					parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are facing wrong way, it's on left side, forward right.");
					return true;
				}
			}
		}
		// If we are actually facing right at the target, we need to turn away from the given wall
		if(wallFrontLeftDistance > 12 && wallFrontRightDistance > 12 && getAreCorrectFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
		{
			// This would be the best one to do of course!
			usedCorrectFacing = true;
			parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are facing waypoint, move forward only.");
			forwardBackwardChosen = FORWARD;
			directionChosen = STRAIGHT;
			return true;
		}
		if(wallFrontLeftDistance > 15 && wallFrontRightDistance > 12 && getAreRightFacing(car.lastCarFacing, waypointAngle, rightAngleLimit, leftAngleLimit))
		{
			parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are on right side of waypoint. Turning left.");
			forwardBackwardChosen = FORWARD;
			directionChosen = LEFT;
			return true;
		}
		
		if(wallFrontRightDistance > 15 && wallFrontLeftDistance > 12)
		{
			parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are on left side of waypoint. Turning right.");
			forwardBackwardChosen = FORWARD;
			directionChosen = RIGHT;
			return true;
		}
		else
		{
			// Move backwards-left
			parentCarAIMain.log("[Navigation AI] navigateDecision() No obstacle. We are on left side of waypoint. Cannot turn right, so going back-left as last resort.");
			forwardBackwardChosen = BACKWARD;
			directionChosen = LEFT;
			return true;
		}
	}

	// Debug command to display all recorded navigation points
	public void displayAllPastPoints()
	{
		parentCarAIMain.log("[Navigation AI] test_DisplayAllPastPoints. pastLocationsArrayIndex [" + pastLocationsArrayIndex + "] pastLocationsArrayOver100 [" + pastLocationsArrayOver100 + "]");
		if(pastLocationsArrayIndex > 0 || pastLocationsArrayOver100 == true)
		{
			// Work on a live image...
			BufferedImage liveImage = parentCarAIMain.camImg.getLiveImage();

			// Add on waypoints for each point
			for(int i = 0; i <= pastLocationsArrayIndex; i++)
			{
				camera.drawBlock(pastLocationsArray[i].x, pastLocationsArray[i].y, liveImage);
				// Draw connecting line
				if(i < pastLocationsArrayIndex)
				{
					camera.drawLine(pastLocationsArray[i], pastLocationsArray[i + 1], liveImage, Color.GREEN);
				}
				//liveImage.setRGB(pastLocationsArray[0].x, pastLocationsArray[0].y, Color.GREEN.getRGB());
			}
			
			// Display
			parentCarAIMain.staticCamImg.image.setImage(liveImage);
		}
	}

	/*************************
		PRIVATE FUNCTIONS 
	*************************/
	
	// Returns true if we are facing nearly directly at the given waypoint
	private boolean getAreCorrectFacing(double carFacing, double waypointAngle, double rightLimit, double leftLimit)
	{
		// Error checking
		if(carFacing < 0 || carFacing > 360 || waypointAngle < 0 || waypointAngle > 360)
		{
			parentCarAIMain.log("[Navigation AI] getAreRightFacing() Error: Inputs invalid!");
			return false;
		}

		// Variance - basically, if we are in this cone, we think we are facing the right direction
		double variance = 15;
		// Are we within this amount of waypointAngle?
		// Note: If for instance, carFacing is 0 and waypointAngle is 10, it would be -10, so we absolute the value. Doesn't actually
		// matter what two angles it is.
		if(Math.abs(carFacing - waypointAngle) <= variance)
		{
			// We are, so it's fine
			return true;
		}
		// Could be waypointAngle or carFacing are either very small or big, overlapping with 0 or 360 together.
		else if(waypointAngle > 360 - variance)
		{
			if(carFacing <= variance - (360 - waypointAngle))
			{
				return true;
			}
		}	
		// And vice-versa
		else if(waypointAngle < 0 + variance)
		{
			if(carFacing > 360 - (variance - waypointAngle))
			{
				return true;
			}
		}
		
		// Else are further out
		return false;
	}
	
	// Returns true if we are facing to the right of the waypoint given, or false if left
	private boolean getAreRightFacing(double carFacing, double waypointAngle, double rightLimit, double leftLimit)
	{
		// Error checking
		if(carFacing < 0 || carFacing > 360 || waypointAngle < 0 || waypointAngle > 360)
		{
			parentCarAIMain.log("[Navigation AI] getAreRightFacing() Error: Inputs invalid!");
			return false;
		}
		// Right facing, we go from waypointAngle to rightLimit
		if(carFacing <= rightLimit && carFacing >= waypointAngle)
		{
			return true;
		}
		// Could still be within the range of rightLimit if it goes over 360
		// Of course, waypointAngle + 180 will NEVER be less then 0. In fact, it will never be less then 180
		else if(rightLimit > 360)
		{
			// We get the overlap
			double overlap = rightLimit - 360;
			// New test
			if(carFacing <= overlap)
			{	
				// Is within 0 and overlap limits. Obviou
				return true;
			}
		}
		// We are otherwise left
		return false;
	}
	
	// Returns true if we are facing the wrong direction to calculate an accurate left / right direction calculation
	private boolean getAreOppositeFacing(double carFacing, double waypointAngle, double rightLimit, double leftLimit)
	{
		// Error checking
		if(carFacing < 0 || carFacing > 360 || waypointAngle < 0 || waypointAngle > 360)
		{
			parentCarAIMain.log("[Navigation AI] getAreRightFacing() Error: Inputs invalid!");
			return false;
		}

		// Variance - basically, if we are in this cone, we think we are facing the wrong direction
		double variance = 70;
		
		// Opposite facing is when they are within the varience of the opposite of the waypoint angle
		// Right Side?
		if(carFacing >= waypointAngle + variance)
		{
			// Within right side?
			if(carFacing <= rightLimit)
			{
				// Wrong way
				return true;
			}
			// Could overlap
			else if(carFacing <= rightLimit - 360)
			{
				// Wrong way
				return true;			
			}
		}
		// Left side?
		if(carFacing <= waypointAngle - variance)
		{
			// Within left side?
			if(carFacing >= leftLimit)
			{
				// Wrong way
				return true;
			}
			// Could overlap (if left limit is <= 0)
			else if(carFacing >= leftLimit + 360)
			{
				// Wrong way
				return true;			
			}
		}	
		// All okay
		return false;
	}
	
	// Return the distance to the nearest obstacle, maximum of150
	private double getDistanceToNearestObstacle(Coordinate car, double angle)
	{
		// This is simple, and inefficient, but still pretty fast.
		// We check, for each point from 0 to 50 at the given angle
		// Sort cos and sin			
		double xMultiply = Math.cos(Math.toRadians(angle));
		double yMultiply = Math.sin(Math.toRadians(angle));
		// Check x and y's
		int x, y, maxDistance = 150;
		for(int i = 1; i <= maxDistance; i++)
		{			
			// Calculate this point to test
			x = Utility.getBoundriedInt(car.x + (int)(i * xMultiply), track.trackMinX, track.trackMaxX);
			y = Utility.getBoundriedInt(car.y + (int)(i * yMultiply), track.trackMinY, track.trackMaxY);

			// Check it
			if(!track.pointWithinTrack(x, y))
			{
				// Return current distance, we've hit a point
				return i;
			}
		}
		// Return max
		return maxDistance;
	}
	
	private void addToPastLocationsArray(Coordinate coord) {
		// Enable button
		parentCarAIMain.displayPastPointsAIButton.setEnabled(true);
		// We start at 0, but the 0th entry is set in the initialRace section above
		pastLocationsArrayIndex++;
		// Can't go over maximum length however (length - 1 is maximum - should be 99)
		if(pastLocationsArrayIndex >= pastLocationsArray.length)
		{
			// Now got over 100 entries, so know we add up by going back to 99 etc.
			pastLocationsArrayOver100 = true;
			pastLocationsArrayIndex = 0;
		}
		// Add it to the array
		pastLocationsArray[pastLocationsArrayIndex] = new Coordinate(coord.x, coord.y);;
	}

	// Gets the last 10 items distance travelled. Will return -1.0 if we have not travelled for 10 recorded seconds yet.
	private double distanceTravelledRecently() {
		// Check...do we have 10 entries? (entries are 0 indexed)
		if(pastLocationsArrayIndex < 10)
		{
			// Might be we add up ones from 99, 98, etc...
			if(pastLocationsArrayOver100 == true)
			{
				// First we add up the values we do have +ve
				double total = Utility.getTotalDistancesFromArray(0, pastLocationsArrayIndex, pastLocationsArray);
				// Get the amount we need to total 0
				int amountLeft = 9 - pastLocationsArrayIndex;
				// Get this amount left
				total += Utility.getTotalDistancesFromArray(99 - amountLeft, 99, pastLocationsArray);

				// Finally, add the value 99 and 0, of course.
				total += Utility.getDistanceBetween(pastLocationsArray[0].x, pastLocationsArray[0].y, pastLocationsArray[99].x, pastLocationsArray[99].y);
				return total;
			}
			else
			{
				// We don't, this check is invalid
				return -1.0;
			}
		}
		// We have at least 10 entries, add them up and return
		return Utility.getTotalDistancesFromArray(pastLocationsArrayIndex - 10, pastLocationsArrayIndex, pastLocationsArray);
	}
}