/**
 * SimCalcs.java
 * Carries out key calculations for simulation
 * Called by SimControl.java for use with Simulation.java
 *
 * Car orientation:
 *
 *					A------B
 *            Rear  	     Front
 *					D------C
 *
 * Created by: Chris Carter
 * Last Modified: 06.05.07
 *
 */
 
import java.lang.*;
import java.awt.*;

public abstract class SimCalcs {

	//Calculate forward movement
	public static SimCar forwards(SimCar car, double thetaInc, double pulseCoeff, double refCoeff){

		int x[] = car.xCoords;
		int y[] = car.yCoords;

		car.thetaX += thetaInc;

		//If angle with x axis is greater than 2PI (i.e. >360 deg, which is one full circle), take 2PI away from it.
		if (car.thetaX >= (2 * Math.PI)){
			car.thetaX -= 2 * Math.PI;
		}

		//Point D is the anchor point for a RH turn, so calculate the movement of point D first
		x[3] += (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.cos(calcThetaR(car.thetaX)));
		y[3] += (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.sin(calcThetaR(car.thetaX)));

		x[0] = (int) Math.round(x[3] + (car.WIDTH * Math.sin(calcThetaR(car.thetaX))));
		y[0] = (int) Math.round(y[3] - (car.WIDTH * Math.cos(calcThetaR(car.thetaX))));

		x[1] = (int) Math.round(x[0] + (car.LENGTH * Math.cos(calcThetaR(car.thetaX))));
		y[1] = (int) Math.round(y[0] + (car.LENGTH * Math.sin(calcThetaR(car.thetaX))));

		x[2] = (int) Math.round(x[3] + (car.LENGTH * Math.cos(calcThetaR(car.thetaX))));
		y[2] = (int) Math.round(y[3] + (car.LENGTH * Math.sin(calcThetaR(car.thetaX))));

		car.xCoords = x;
		car.yCoords = y;

		//Update graphical representation of car
		car.carPoly = new SimCar2D(car.xCoords, car.yCoords);

		return car;

	}

	//Calculate movement for RH turn
	public static SimCar turnRight(SimCar car, double thetaInc, double pulseCoeff, double refCoeff){

		int x[] = car.xCoords;
		int y[] = car.yCoords;

		car.thetaX += thetaInc;

		//If angle with x axis is greater than 2PI (i.e. >360 deg, which is one full circle), take 2PI away from it.
		if (car.thetaX >= (2 * Math.PI)){
			car.thetaX -= 2 * Math.PI;
		}

		//Point D is the anchor point for a RH turn, so calculate the movement of point D first
		x[3] += (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.cos(calcThetaR(car.thetaX)));
		y[3] += (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.sin(calcThetaR(car.thetaX)));

		x[0] = (int) Math.round(x[3] + (car.WIDTH * Math.sin(calcThetaR(car.thetaX))));
		y[0] = (int) Math.round(y[3] - (car.WIDTH * Math.cos(calcThetaR(car.thetaX))));

		x[1] = (int) Math.round(x[0] + (car.LENGTH * Math.cos(calcThetaR(car.thetaX))));
		y[1] = (int) Math.round(y[0] + (car.LENGTH * Math.sin(calcThetaR(car.thetaX))));

		x[2] = (int) Math.round(x[3] + (car.LENGTH * Math.cos(calcThetaR(car.thetaX))));
		y[2] = (int) Math.round(y[3] + (car.LENGTH * Math.sin(calcThetaR(car.thetaX))));

		car.xCoords = x;
		car.yCoords = y;

		//Update graphical representation of car
		car.carPoly = new SimCar2D(car.xCoords, car.yCoords);

		return car;

	}

	//Calculate movement for LH turn
	public static SimCar turnLeft(SimCar car, double thetaDec, double pulseCoeff, double refCoeff){

		int x[] = car.xCoords;
		int y[] = car.yCoords;

		car.thetaX -= thetaDec;

		//If angle with x axis is less than -2PI (i.e. < -360 deg, which is one full circle), add 2PI to it.
		if(car.thetaX <= (-2 * Math.PI)){
			car.thetaX += 2 * Math.PI;
		}

		//Point A is the anchor point for a LH turn, so calculate the movement of point A first
		x[0] += (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.cos(calcThetaL(car.thetaX)));
		y[0] -= (int) Math.round(car.REFRATE * pulseCoeff * refCoeff * Math.sin(calcThetaL(car.thetaX)));

		x[1] = (int) Math.round(x[0] + (car.LENGTH * Math.cos(calcThetaL(car.thetaX))));
		y[1] = (int) Math.round(y[0] - (car.LENGTH * Math.sin(calcThetaL(car.thetaX))));

		x[2] = (int) Math.round(x[1] + (car.WIDTH * Math.sin(calcThetaL(car.thetaX))));
		y[2] = (int) Math.round(y[1] + (car.WIDTH * Math.cos(calcThetaL(car.thetaX))));

		x[3] = (int) Math.round(x[0] + (car.WIDTH * Math.sin(calcThetaL(car.thetaX))));
		y[3] = (int) Math.round(y[0] + (car.WIDTH * Math.cos(calcThetaL(car.thetaX))));

		car.xCoords = x;
		car.yCoords = y;

		//Update graphical representation of car
		car.carPoly = new SimCar2D(car.xCoords, car.yCoords);

		return car;

	}

	//Returns correct theta for use in calculations for RH turns
	public static double calcThetaR(double theta){

		double result;
		if(theta < 0){
			result = theta + (Math.PI*2);
		} else {
			result = theta;
		}
		return result;

	}

	//Returns correct theta for use in calculations for LH turns
	private static double calcThetaL(double theta){

		double result;
		if(theta < 0){
			result = Math.abs(theta);
		} else {
			result = Math.abs(theta - (Math.PI*2));
		}
		return result;

	}

	//public static void accelerate(int max){
	//}

}