import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;

/**
 * SimDraw.java
 *
 * Main simulation interface class
 * Draws simulated track and car to scale in current position, for use with Simulation.java
 * Scale is 1px : 2mm, with track dimensions rounded down and car/wall dimensions rounded up for lower chance of error
 * Controls the simulation upon receiving events from Simulation.java
 *
 * Car orientation:
 *
 *					A------B
 *            Rear  	     Front
 *					D------C
 *
 * Therefore point A is 1st referenced point.
 *
 * The 1px : 2mm doesn't fit now, and is scaled down so the GUI isn't as cluttered since the monitor is smaller.
 * But, it does what it says; draws a simulation of the track, ie; Walls, Corners, Obstacles.
 * Notably has all the controls and other GUI elements in this class - including the code for when
 * the simulation is run (ie; when you can click "Stop" during the simulation).
 * THIS NOW IS CHANGED:
 * Should be ONLY THE SIMULATION CODE, and have a function that can run as part of a threaded part of the program, replacing
 * the need for threads in this function.
 * No GUI code here either.
 * This does have the direct GUI-stored information like "clicker number" but none of it is setup here.
 *
 * Renamed to SimDraw, when previously was DrawSim.
 * 
 * Created by: Chris Carter
 * Modified by: Andrew Armstrong
 * Last Modified: 07.04.08
 *
 */

public class SimDraw extends JPanel implements Runnable {

	// Cleaned up variables only used for simulation work
	int initial_x, initial_y;// The position on the diagram we want the car to start at.
	int pos = 0;
	boolean paused;// Paused simulation or not
	int clockRate;// Speed of the simulation
	boolean driveCar;// ???
	boolean stopOnCrash;// Do we stop the simulation when the program detects a crash?

	// Thread to run the simulation on
	Thread runner;
	
	SimCar car;

	CarAIMain parentCarAIMain; // "Parent", GUI stuff like messages need this.
	SimActionHandler actHandler; // The action read from the text file, this class reads more stuff

	SimCarControl carControl = new SimCarControl();

	SimAction currentAction; // The *current* action read (basically a number)

	String coordDetails;

	private final static long serialVersionUID = 1L;

	//Identify areas of track and whether they are driveable
/*
 	//OuterWalls
	SimTrackArea outerWalls = new SimTrackArea(30,10,1220,607, false);
	//Track Floor i.e. all black visible to camera (added as black rectangle 20x20 px smaller than Outer Walls rectangle)
	SimTrackArea trackFloor = new SimTrackArea(40,20,1200,587, true);
	//Small Wall that sticks out at 'top' of track
	SimTrackArea smallWall = new SimTrackArea(630,20,20,150, false);
	//Central inner wall across middle
	SimTrackArea centreWall = new SimTrackArea(170,305,940,19, false);
*/
	// Halved version
	//OuterWalls
	SimTrackArea outerWalls = new SimTrackArea(15,5,610,303, false);

	//Track Floor i.e. all black visible to camera (added as black rectangle 20x20 px smaller than Outer Walls rectangle)
	SimTrackArea trackFloor = new SimTrackArea(20,10,600,293, true);

	//Small Wall that sticks out at 'top' of track
	SimTrackArea smallWall = new SimTrackArea(315,10,10,75, false);

	//Central inner wall across middle
	SimTrackArea centreWall = new SimTrackArea(85,152,470,8, false);

	SimTrackArea[] innerTrack = {trackFloor, smallWall, centreWall};
	
	//Constructer establishing start position, and initialises simulation
	public SimDraw(CarAIMain passedCarAIMain){

		// Parent for logging
		parentCarAIMain = passedCarAIMain;

		// Half input sizes (for the car starting location) down from 547 and 218 to 274 114
		initial_x = 274;
		initial_y = 114;

		//Generate a new car opbject
		car = new SimCar(initial_x, initial_y, 0);
		coordDetails = "Initial Pos: A: ("+(car.xCoords[0]-30)+","+(car.yCoords[0]-10)+") B: ("+(car.xCoords[1]-30)+","
						+(car.yCoords[1]-10)+") C: ("+(car.xCoords[2]-30)+","+(car.yCoords[2]-10)+") D: ("+(car.xCoords[3]-30)+","
						+(car.yCoords[3]-10)+")\n";

		parentCarAIMain.logCoords(coordDetails);
		String[] endLoc = {"Top", "Right"};

		//Generate new ActionHandler and prepare first action
		actHandler = new SimActionHandler(endLoc, 10, car, parentCarAIMain);
		currentAction = actHandler.nextAction();

		//Determine initial sim options
		clockRate = parentCarAIMain.clockSlider.getValue();
		driveCar = parentCarAIMain.driveCarCB.isSelected() ? true : false;
		stopOnCrash = parentCarAIMain.stopCrashCB.isSelected() ? true : false;
		
		// Detailed logging turns this on
		parentCarAIMain.logDetail("Initial angle with x axis is: " + car.thetaX +"\n");
	}

	//Do this when Start/Resume Simulation is clicked
	void startPlaying() {
		//Start simulation
		runner = new Thread(this);
		runner.start();
		paused = false;

		//Enable/disable appropriate controls
		parentCarAIMain.stopButton.setEnabled(true);
		parentCarAIMain.startButton.setEnabled(false);
		parentCarAIMain.resetButton.setEnabled(false);
		parentCarAIMain.driveCarCB.setEnabled(false);
		if(driveCar){
			parentCarAIMain.clockSlider.setEnabled(false);
		}
	}

	//Do this when Stop Simulation is clicked
	void stopPlaying() {
		//Turn off data pins of parallel port
		carControl.close(2);
		parentCarAIMain.log("@ pos {" + pos + "} " + carControl.exitMsg);

		// Trigger termination of sim
		paused = true;

		// Enable/disable appropriate controls
		parentCarAIMain.stopButton.setEnabled(false);
		parentCarAIMain.startButton.setText("Resume Simulation");
	}

	// Do this when Reset is clicked
	void resetSim(){
		if(parentCarAIMain.saveLog){
			// create a new log file at next input, don't stop logging however!
			parentCarAIMain.closeLog();
		}
		// Destroy logs
		parentCarAIMain.log("", true);
		parentCarAIMain.logCoords("", true);
		// Reset other things
		pos = 0;
		car = new SimCar(initial_x, initial_y, 0);
		coordDetails = "Initial Pos: A: ("+(car.xCoords[0]-30)+","+(car.yCoords[0]-10)+") B: ("+(car.xCoords[1]-30)+","
						+(car.yCoords[1]-10)+") C: ("+(car.xCoords[2]-30)+","+(car.yCoords[2]-10)+") D: ("+(car.xCoords[3]-30)+","
						+(car.yCoords[3]-10)+")\n";
						
					
						
		parentCarAIMain.logCoords(coordDetails);
		parentCarAIMain.logDetail("Initial angle with x axis is: " + car.thetaX +"\n");

		String[] endLoc = {"Top", "Right"};
		actHandler = new SimActionHandler(endLoc, 10, car, parentCarAIMain);
		currentAction = actHandler.nextAction();
		paused = false;
		repaint();
		if(parentCarAIMain.startButton.getText() == "Resume Simulation"){
			parentCarAIMain.startButton.setText("Start Simulation");
		}

		//Enable/disable appropriate controls
		parentCarAIMain.resetButton.setEnabled(false);
		parentCarAIMain.startButton.setEnabled(true);
		parentCarAIMain.startButton.requestFocus(true);
		parentCarAIMain.clockSlider.setEnabled(true);
		if(clockRate == 1){
			parentCarAIMain.driveCarCB.setEnabled(true);
		}
	}

	//Do this when Simulation reaches a natural (i.e. non-user interrupted) conclusion
	void endOfSim() {
		runner = null;
		parentCarAIMain.resetButton.setEnabled(true);
		parentCarAIMain.stopButton.setEnabled(false);
	}

	//Draw current position on screen
	public void paintComponent(Graphics comp) {
		super.paintComponent(comp);

		//Draw track
		Graphics2D gfxOuterWalls = (Graphics2D) comp;
		gfxOuterWalls.setColor(Color.white);
		gfxOuterWalls.fill(outerWalls);

		Graphics2D gfxTrackFloor = (Graphics2D) comp;
		gfxTrackFloor.setColor(Color.black);
		gfxTrackFloor.fill(trackFloor);

		Graphics2D gfxSmallWall = (Graphics2D) comp;
		gfxSmallWall.setColor(Color.white);
		gfxSmallWall.fill(smallWall);

		Graphics2D gfxCentreWall = (Graphics2D) comp;
		gfxCentreWall.setColor(Color.white);
		gfxCentreWall.fill(centreWall);

		//Draw the car's current position on the track
		Graphics2D gfxCar = (Graphics2D) comp;
		gfxCar.setRenderingHint(RenderingHints.KEY_ANTIALIASING,RenderingHints.VALUE_ANTIALIAS_ON); //Smooth diagonal lines when turning
		gfxCar.setColor(car.COLOR);
		gfxCar.fill(car.carPoly);

	}

	//Simulate car movement
	// This is the thread which runs to make the simulation do things over time (without locking up the GUI).
	public void run() {

		//Determine which Thread is currently running
		Thread thisThread = Thread.currentThread();

		//Actions for main sim thread
		while (runner == thisThread){
					
			//Lag time constants
			int lagTimeStrght = 35;
			int lagTimeTrn = 90;

			pos++;

			//Stop sim if 'Stop Simulation' button has been clicked
			if(paused){
				try{
					Thread.sleep(1000);
				} catch (InterruptedException e){
					parentCarAIMain.log("Error sleeping thread for pausing. Error: " + e.toString());
				}

				//Enable/disable appropriate controls
				parentCarAIMain.startButton.setEnabled(true);
				parentCarAIMain.resetButton.setEnabled(true);
				parentCarAIMain.startButton.requestFocus(true);

				//Stop the sim
				runner = null; // Why not call endOfSim()? it's only paused!
				if(clockRate == 1){
					parentCarAIMain.driveCarCB.setEnabled(true);
				}
				parentCarAIMain.clockSlider.setEnabled(true);
				parentCarAIMain.log("Car stopped @ pos {" + pos + "}");
				parentCarAIMain.saveLogCB.setEnabled(true);
			}

			//If there are no more instructions to execute, stop the sim
			else if (currentAction.noMoreInsts){
				carControl.close(5);
				parentCarAIMain.log("@ pos {" + pos + "} " + carControl.exitMsg);
				parentCarAIMain.logDetail("Final angle with x axis is: " + car.thetaX);
				endOfSim();
			}

			//When car has gone as far as required, stop the sim
			else if (actHandler.getComplete()){
				carControl.close(1);
				parentCarAIMain.log("@ pos {" + pos + "} " + carControl.exitMsg);
				parentCarAIMain.logDetail("Final angle with x axis is: " + car.thetaX);
				endOfSim();
			}

			//Stop sim and report an error if bad instruction detected
			else if (currentAction.error){
				carControl.close(-1);
				parentCarAIMain.log("@ pos {" + pos + "} " + carControl.exitMsg);
				parentCarAIMain.logDetail("Angle with x axis when error occured: " + car.thetaX);
				endOfSim();
			}

			//If all tests are cleared, continue/begin next action
			else {
				if(actHandler.actCounter == 0){// i.e. if this is the first action
					actHandler.beginNextAction(car, currentAction, driveCar);
					parentCarAIMain.logDetail(getNewActInfo());
				}
				else if(currentAction.complete){// i.e. if new action required
					currentAction = actHandler.nextAction();
					actHandler.beginNextAction(car, currentAction, driveCar);
					parentCarAIMain.logDetail(getNewActInfo());
				}
				else {// continue current action
					actHandler.continueAction(car, currentAction, driveCar);
				}

				// Update display to show new position
				repaint();

			}
			coordDetails = "{"+pos+"} A: ("+(car.xCoords[0]-30)+","+(car.yCoords[0]-10)+") B: ("+(car.xCoords[1]-30)+","
							 +(car.yCoords[1]-10)+") C: ("+(car.xCoords[2]-30)+","+(car.yCoords[2]-10)+") D: ("+(car.xCoords[3]-30)
							 +","+(car.yCoords[3]-10)+")"; 
		
							
			parentCarAIMain.logCoords(coordDetails);
			parentCarAIMain.logDetail("@ pos {" + pos + "} Angle(x): " + car.thetaX + " " + actHandler.getPowerInfo());

			// Sleep for length of time, determined by required pulse
			// N.B. NEEDS EDITING ONCE REVERSE ANIMATIONS ETC. ARE CODED
			if(currentAction.type == 1){//i.e. straight

				try{
					Thread.sleep((int) Math.round(((100 * currentAction.pulseCoeff) + lagTimeStrght) / clockRate));
				} catch(InterruptedException e) {
					parentCarAIMain.log("Error sleeping thread for required action delay time. Error: " + e.toString());
				}

			}
			else {//i.e. turning

				try{
					Thread.sleep((int) Math.round(((100 * currentAction.pulseCoeff) + lagTimeTrn) / clockRate));
				} catch(InterruptedException e) {
					parentCarAIMain.log("Error sleeping thread for required action delay time. Error: " + e.toString());
				}

			}

			// Test for collision with wall
			if(SimTestCrash.testPos(innerTrack, parentCarAIMain, car)) {
				if(stopOnCrash){
					carControl.close(3);
				}
				else {
					carControl.close(4);
				}
				parentCarAIMain.log("@ pos {" + pos + "} " + carControl.exitMsg);
				parentCarAIMain.logDetail("Angle with x axis when collision occured: " + car.thetaX);
				if(stopOnCrash){
					endOfSim();
				}
			}
		}
	}

	//Returns details on the new action to be executed. Called if detailed feedback information is requested
	private String getNewActInfo(){

		String dir, steer;
		switch(currentAction.type){
			case 1:	 dir = "FW";
					 steer = "STRGHT";
					 break;

			case 2:	 dir = "FW";
					 steer = "R";
					 break;

			case 3:	 dir = "FW";
					 steer = "L";
					 break;

			case 4:	 dir = "BW";
					 steer = "STRGHT";
					 break;

			case 5:	 dir = "BW";
					 steer = "R";
					 break;

			case 6:	 dir = "BW";
					 steer = "L";
					 break;

			default: dir = "END";
					 steer = "END";
					 break;
		}
		return "New Action (" + dir + "," + steer + ") pulses: " + currentAction.pulses + " coeff: " + currentAction.pulseCoeff;
	}

	// AA: saves checking lots of subclass stuff - is the car in the initial position basically...
	public boolean simStarted(){
		// Returns FALSE if we have not started
		if((car.xCoords[0] == initial_x) && (car.yCoords[0] == initial_y)) return false;
		// Else return TRUE
		return true;
	}
	
	// Easy check to see if the simulation is running (used to test some GUI things if altered mid-run)
	public boolean simRunning() {
		// Return true if we have a valid runner thread
		return (runner != null);
	}
	
	// AA: This produces the initial logging (or clears the log) as needed
	public void simLoggingReset() {
		if(parentCarAIMain.detInfo){
			parentCarAIMain.logDetail("Initial angle with x axis is: " + car.thetaX);
		}
		else{
			// AA: Clear log 
			parentCarAIMain.log("", true);
		}
	}

}