/**
 * Track.java
 *
 * This class works with the various track components to sort out:
 * - Track locations:
 *		- Walls
 *		- Centre barrier
 *		- Chicane
 * - If a point, (X,Y) is in a valid track location or not (this is used for testing for crashes, and so forth)
 * - The waypoints on the track, and the current one the car needs to reach
 * 
 * This also stores all the readable information about track limits. Some is publicly available to get, since
 * it allows an easy way to lower the CPU overhead of checking *every* pixel in an image for something.
 *
 * Created By: Andrew Armstrong
 * Last Modified: 04/05/08
 */

import java.util.*;
import java.awt.*;
import java.awt.image.*;

public class Track {

	/*************************
		Variables
	*************************/

	// Variables for what we alter the final image to be - empty is black, full is green.
	public static final int EMPTY_REGION = 0x000000;
	public static final int FULL_REGION  = 0x00FF00;

	public static final int MATCH_EMPTY_REGION = EMPTY_REGION - 16777216; // For getRGB, magic number removed!
	public static final int MATCH_FULL_REGION  = FULL_REGION - 16777216; // For getRGB, magic number removed!

	// Track limits. These are the absolute edge of the track, and are really a good idea to do first!
	// Anything outside this "doesn't exist" essentially.
	// Defaults are from a camera image size - note, this might need altering if camera changes, etc.
	int trackMinX = 0, trackMinY = 0, trackMaxX = 320, trackMaxY = 240;

	// Internal wall limits. These differ from the above since we can assume that once a identified wall pixel *could* be hit, 
	// it will have crashed. These are not used for cutting down CPU time, and are only interfaced with the "in or out of track" classes.
	private int wallMinX = 0, wallMinY = 0, wallMaxX = 320, wallMaxY = 240;
	
	// Information on the centre barrier
	// Information applies from the assumed "top left" of the barrier, as a retangle.
	private int barrierX = 0, barrierY = 0, barrierSizeX = 0, barrierSizeY = 0;

	// Information about the chicane
	// This assumes less information then the barrier, since we know the upper limit of the chicane is the "trackMinY" value, which we take as the Y value for it.
	private int chicaneX = 0, chicaneSizeX = 0, chicaneSizeY = 0;
	
	// Waypoints are stored in a simple array, which we need to navigate around, and is setup below.
	// The array is setup as 2D co-ordinates (same as other co-ordinates in the system)
	private Coordinate[] waypointsArray; // Not initalised yet - no knowing how many points we need!

	// Current waypoint we are going towards. 0 index of course.
	private int waypointCurrent = 0;

	// For logging and ease of use
	private CarAIMain parentCarAIMain;
	private Camera camera;
	
	/*************************
		Constructor
	*************************/

	public Track(CarAIMain passedCarAIMain, Camera passedCamera)
	{
		passedCarAIMain.logDetail("[Track] Constructor: Starting");
		parentCarAIMain = passedCarAIMain;
		this.camera = passedCamera;
		passedCarAIMain.logDetail("[Track] Constructor: Done");
	}

	/*************************
		PUBLIC FUNCTIONS 
	*************************/

	// Find the "Walls" - max and min X and Y points for use in anything to do with the track.
	public boolean findTrackArea(BufferedImage thresholdImage, BufferedImage debugImage, Car car) {
		// Camera needs to be on!
		if(!camera.getCameraOn()) {
			parentCarAIMain.log("[Track] findTrackArea() - Error, camera not on.");
			return false;
		}
		// Null input?
		if(thresholdImage == null || thresholdImage.getWidth() <= 0) {
			parentCarAIMain.log("[Track] findTrackArea() - Error, no valid input image");
			return false;
		}

		// Initial max and min X and Y
		int x = 0, minX = 0, maxX = thresholdImage.getWidth();
		int y = 0, minY = 0, maxY = thresholdImage.getHeight();

		// Error checking (esp. for /0 errors and accidents)
		if(maxX <= 0 || maxY <= 0) {
			parentCarAIMain.log("[Track] findTrackArea() - Error, no image width or height.");
			return false;
		}
		// Else log, debug the sizes so can see improvements
		int areaX = maxX;
		int areaY = maxY;
		int pixels = maxX * maxY;
		parentCarAIMain.logDetail("[Track] findTrackArea() - Searching for track limits in image: minX [" + minX + "] MaxX [" + maxX + "] MinY [" + minY + "] MaxY [" + maxY + "]\nArea: (" + areaX + " x " + areaY + "). Pixels [" + pixels + "]");
		
		// Centre points
		int centreX = maxX / 2;
		int centreY = maxY / 2;

		// We now do this very simply; we scanline the image (from the centre!) until we find the first totally black line
		// Pan and Scan (the top and bottom of the image - 9 pixels worth - is always black) means we have to work to get the Y values first.
		// Since we only do this once per "track capture" there really is no need to optimise the loops much. This is the start of optimisation really!
		
		// Min Y is towards the top of the image, since "top is 0" in Y.
		for(y = centreY; y >= minY; y--) {
			// Check for this line
			if(scanHorizontalLineForColor(y, minX, maxX, MATCH_EMPTY_REGION, thresholdImage)) {
				// Take the first one
				minY = y;
				break;
			}
		}
		// Max Y is from the centre up to MaxY
		for(y = centreY; y < maxY; y++) {
			// Check for this line
			if(scanHorizontalLineForColor(y, minX, maxX, MATCH_EMPTY_REGION, thresholdImage)) {
				// Take the first one
				maxY = y;
				break;
			}
		}
		// Min X is from the centre going to the left - now using new minY and maxY values
		for(x = centreX; x >= minX; x--) {
			// Check for this line 
			if(scanVerticalLineForColor(x, minY, maxY, MATCH_EMPTY_REGION, thresholdImage)) {
				// Take the first one
				minX = x;
				break;
			}
		}
		// Max X is from the centre going to the right - now using new minY and maxY values
		for(x = centreX; x < maxX; x++) {
			// Check for this line
			if(scanVerticalLineForColor(x, minY, maxY, MATCH_EMPTY_REGION, thresholdImage)) {
				// Take the first one
				maxX = x;
				break;
			}
		}
		// Report the findings
		areaX = maxX - minX;
		areaY = maxY - minY;
		pixels = areaX * areaY;
		parentCarAIMain.log("[Track] findTrackArea() - Finished: minX [" + minX + "] MaxX [" + maxX + "] MinY [" + minY + "] MaxY [" + maxY + "]\nOptimised area: (" + areaX + " x " + areaY + "). Pixels [" + pixels + "]");

		// Set the values now for error checking (important to keep this at the end in case I want to alter how the function works later)
		if(minX == 0) {
			parentCarAIMain.log("[Track] findTrackArea() - Error: found minimum X equals the image minimum, 0. Check image quality.");
			return false;
		}
		if(minY == 0) {
			parentCarAIMain.log("[Track] findTrackArea() - Error: found minimum Y equals the image minimum, 0. Check image quality.");
			return false;
		}
		if(maxX == thresholdImage.getWidth()) {
			parentCarAIMain.log("[Track] findTrackArea() - Error: found maximum X equals the image maximum: " + thresholdImage.getWidth() + ". Check image quality.");
			return false;
		}
		if(maxX == thresholdImage.getHeight()) {
			parentCarAIMain.log("[Track] findTrackArea() - Error: found maximum Y equals the image maximum: " + thresholdImage.getHeight() + ". Check image quality.");
			return false;
		}
		// If it got this far it worked, no errors. Set variables
		trackMinX = minX;
		trackMaxX = maxX;
		trackMinY = minY;
		trackMaxY = maxY;
		// Set on car too (copies the variables essentially)
		car.setTrackSize(minX, maxX, minY, maxY);
			
		//*
		// Debug drawing
		camera.drawHorizontalLine(minY, debugImage);
		camera.drawHorizontalLine(maxY, debugImage);
		camera.drawVerticalLine(minX, debugImage);
		camera.drawVerticalLine(maxX, debugImage);		
		// Display it
		//parentCarAIMain.staticCamImg.image.setImage(thresholdImage);
		//*/
		return true;
	}
	
	// Find more "true" walls, described below.
	public boolean findWalls(BufferedImage thresholdImage, BufferedImage debugImage) {
		// Errors
		if(thresholdImage == null || thresholdImage.getWidth() <= 0) {
			parentCarAIMain.log("[Track] findWalls: Error, invalid input image");
			return false;
		}
		// Is the image big enough?
		if(thresholdImage.getWidth() < trackMaxX || thresholdImage.getHeight() < trackMaxY) {
			parentCarAIMain.log("[Track] findWalls: Error, track variables are larger then input image");
			return false;
		}
		// Chicane?
		if(chicaneSizeY == 0) {
			parentCarAIMain.log("[Track] findWalls: Error, no chicane!");
			return false;
		}

		// A wall is not the "most near black line" but is instead, going the opposite way, the more internal track line.
		// This is done because the track walls are never EVER detected cleanly.
		// The problem with this is that it is difficult to judge, since we only know the track extremes, we don't know where to stop wall analysis...ie:
		// Top of track goes along the "minX" line, and moves down scanning. We don't know the internal minY and maxY variable, of course! We therefore work on the basis
		// of "% track is needed", defined here:
		int minimumFullPercent = 80;

		// Scan the lines...
		int x = 0, y = 0;

		// Default them to the "track" variables, which will do in a pinch, but we'll report the differences at the end.
		wallMinX = trackMinX;
		wallMinY = trackMinY;
		wallMaxX = trackMaxX;
		wallMaxY = trackMaxY;

		// Start with the top line, wallMinY
		// We work DOWN from the trackMinY, ie; adding to it
		for(y = trackMinY; y < trackMaxY; y++) {
			// Check for this line
			// We miss out the gap by the chicane!
			if(scanHorizontalLineForColor(y, trackMinX, trackMaxX, MATCH_FULL_REGION, thresholdImage, minimumFullPercent, chicaneX, chicaneX + chicaneSizeY)) {
				// Take the first one
				wallMinY = y;
				break;
			}
		}
		
		// Scan the bottom line, wallMaxY
		for(y = trackMaxY; y > trackMinY; y--) {
			// Check for this line (no chicane)
			if(scanHorizontalLineForColor(y, trackMinX, trackMaxX, MATCH_FULL_REGION, thresholdImage, minimumFullPercent)) {
				// Take the first one
				wallMaxY = y;
				break;
			}
		}
		
		// Now we have minimum and maximum X values, we can do X more accurately
		// Do left side, wallMinX
		for(x = trackMinX; x < trackMaxX; x++) {
			// Check for this line (no chicane)
			if(scanVerticalLineForColor(x, wallMinY, wallMaxY, MATCH_FULL_REGION, thresholdImage, minimumFullPercent)) {
				// Take the first one
				wallMinX = x;
				break;
			}
		}
		// Do right side, wallMaxX
		for(x = trackMaxX; x > trackMinX; x--) {
			// Check for this line (no chicane)
			if(scanVerticalLineForColor(x, wallMinY, wallMaxY, MATCH_FULL_REGION, thresholdImage, minimumFullPercent)) {
				// Take the first one
				wallMaxX = x;
				break;
			}
		}
		
		// Report differences	
		parentCarAIMain.log("[Track] findWalls result (cannot have error). trackMinX/wallMinX [" + trackMinX + "/" + wallMinX + "] minY [" + trackMinY + "/" + wallMinY + "] maxX [" + trackMaxX + "/" + wallMaxX + "] MaxY [" + trackMaxY + "/" + wallMaxY + "]");

		// Debug drawing
		//*
		camera.drawRectangle(wallMinX, wallMinY, wallMaxX - wallMinX, wallMaxY - wallMinY, debugImage);
		//*/
		return true;
	}
	
	// Find the "chicane" in the track
	public boolean findChicane(BufferedImage thresholdImage, BufferedImage debugImage) {
		// Errors
		if(thresholdImage == null || thresholdImage.getWidth() <= 0) {
			parentCarAIMain.log("[Track] findChicane: Error, invalid input image");
			return false;
		}
		// Is the image big enough?
		if(thresholdImage.getWidth() < trackMaxX || thresholdImage.getHeight() < trackMaxY) {
			parentCarAIMain.log("[Track] findChicane: Error, track variables are larger then input image");
			return false;
		}
		// Reset chicane data
		chicaneX = 0;
		chicaneSizeX = 0;
		chicaneSizeY = 0;

		// We look for a chicane like so:
		// - We know the centre of the "track" by trackMinX/MinY/MaxX/MaxY
		// - The chicane should be roughly in the centre, found below the centre of the top wall
		// 	- therefore, we search a line near the centre, with it starting 1/5 of the way down, which should hit it
		//	- Missing is a big no!
		// 	- We also want to find the biggest width, so we do it here (if this doesn't work, do it once we know the length)
		// - Once we have a point on it, we find the furthest point scanning down.
		// Steps below!
		int x = 0, y = 0, centreX, totalPoints, coordinateTotal, avgX;
		boolean breakLoop = false, breakInternalLoop = false;
		
		// - We know the centre of the "track" by trackMinX/MinY/MaxX/MaxY
		int startingPointX = trackMinX + ((trackMaxX - trackMinX) / 2);// Middle of X
		int startingPointY = trackMinY + ((trackMaxY - trackMinY) / 5);// Top line + 1/5 of distance 

		// Log
		parentCarAIMain.logDetail("[Track] findChicane, start. startingPointX [" + startingPointX + "] startingPointY [" + startingPointY + "]");
		
		// - The chicane should be roughly in the centre, found below the centre of the top wall
		// 	- therefore, we search a line near the centre, with it starting 1/5 of the way down, which should hit it
		//	- Missing is a big no!
		// 	- We also want to find the biggest width, so we do it here (if this doesn't work, do it once we know the length)
		// Y is up and down. Start 1/5 of the way down, go further down, looking for a valid Y co-ordinate
		for(y = startingPointY; y < startingPointY + 20 && breakLoop == false; y++) {
			// Scan left/right 10 each way
			for(x = startingPointX - 10; x < startingPointX + 10 && breakLoop == false; x++) {
				// Check for this line (no chicane)
				if(thresholdImage.getRGB(x, y) == MATCH_EMPTY_REGION) {
					// Take the first one as a valid part of the chicane.
					// The problem: this might be an "odd" point...so...we need to get the average of the next points until we hit "track"
					// Sort this here
					totalPoints = 1;
					coordinateTotal = x;
					breakInternalLoop = false;
					// Has something to stop it after 10 max, and start from x+1
					for(avgX = x + 1; avgX <= x + 11 && breakInternalLoop == false; avgX++) {
						// Empty?
						if(thresholdImage.getRGB(avgX, y) == MATCH_EMPTY_REGION) {
							totalPoints++;
							coordinateTotal += avgX;
						}
						else
						{
							// Stop if we reach a "full" area.
							breakInternalLoop = true;
						}
					}
					// Do need at LEAST 3 points
					if(totalPoints >= 3)
					{
						// Temporary assignment
						centreX = coordinateTotal / totalPoints;
						// This is proper assignment
						chicaneSizeX = totalPoints;
						chicaneX = x;
						breakLoop = true;
					}
				}
			}
		}

		// Error checking
		if(chicaneX == 0)
		{
			// Error
			parentCarAIMain.log("[Track] findChicane Error. Not found chicane area - chicaneX is zero");
			return false;
		}
		
		// - Once we have a point on it, we find the furthest point scanning down that line, this being the length...
		for(y = startingPointY; y < startingPointY + 40; y++) {
			// Check point
			if(thresholdImage.getRGB(x, y) == MATCH_EMPTY_REGION) {
				// Go until we have a non-empty region, so assign this
				chicaneSizeY = y - trackMinY;
			}
			else
			{
				// End, use what's assigned
				break;
			}
		}
		// Report size
		parentCarAIMain.log("[Track] findChicane result. chicaneX [" + chicaneX + "] chicaneSizeX [" + chicaneSizeX + "] chicaneSizeY [" + chicaneSizeY + "]");
		// Debug drawing
		//*
		camera.drawRectangle(chicaneX, trackMinY, chicaneSizeX, chicaneSizeY, debugImage);
		//*/	
		return true;	
	}
	
	// Find the centre barrier
	// Needs the chicane found first
	public boolean findBarrier(BufferedImage thresholdImage, BufferedImage debugImage) {
		// Errors
		if(thresholdImage == null || thresholdImage.getWidth() <= 0) {
			parentCarAIMain.log("[Track] findBarrier: Error, invalid input image");
			return false;
		}
		// Is the image big enough?
		if(thresholdImage.getWidth() < trackMaxX || thresholdImage.getHeight() < trackMaxY) {
			parentCarAIMain.log("[Track] findBarrier: Error, track variables are larger then input image");
			return false;
		}
		// Chicane?
		if(chicaneSizeY == 0) {
			parentCarAIMain.log("[Track] findBarrier: Error, no chicane!");
			return false;
		}
		
		// Default back the barrier stuff
		// Information applies from the assumed "top left" of the barrier, as a retangle.
		barrierX = 0;
		barrierY = trackMaxY; // Set to a high amount since we want a minimal value
		barrierSizeX = 0;
		barrierSizeY = 0;
		
		// We search for it in these stages:
		// - Assume is realtively near the centre of the track limits (we can't say it "always" is however). Assume the starting Y value is the chicane's lowest point
		// 	- Search down for the closest Y it appears
		//	- Find limits (for the width, and minY values) and average the centre
		// - Find X limits, by using the centre Y points, and find the extent left, then right, assign to min and max X respectively
		// Found it!
		
		// - Assume is realtively near the centre of the track limits (we can't say it "always" is however). Assume the starting Y value is the chicane's lowest point
		int initialX = trackMinX + ((trackMaxX - trackMinX) / 2);// Middle of X
		int initialY = trackMinY + chicaneSizeY + 1; // Start just below the chicane 
			
		int x = 0, y = 0, centreY = initialY, totalPoints, coordinateTotal, avgY, counter = 0;
		boolean breakLoop = false, breakInternalLoop = false;
		
		// 	- Search down for the closest Y it appears
		for(y = initialY; y <= trackMaxY - 10; y++)
		{
			// Found it?
			if(thresholdImage.getRGB(initialX, y) == MATCH_EMPTY_REGION) {
				// Take the first one as a valid part of the barrier.
				// The problem: this might be an "odd" point...so...we need to get the average of the next points until we hit "track"
				// Sort this here
				totalPoints = 1;
				coordinateTotal = y;
				breakInternalLoop = false;
				// Has something to stop it after 10 max, and start from y+1
				for(avgY = y + 1; avgY <= y + 11 && breakInternalLoop == false; avgY++) {
					// Empty?
					if(thresholdImage.getRGB(initialX, avgY) == MATCH_EMPTY_REGION) {
						totalPoints++;
						coordinateTotal += avgY;
					}
					else
					{
						// Stop if we reach a "full" area.
						breakInternalLoop = true;
					}
				}
				// Do need at LEAST 3 points
				if(totalPoints >= 3)
				{
					// Temporary assignment
					centreY = coordinateTotal / totalPoints;
					breakLoop = true;
				}
			}
		}
		// Should have barrierY defined, and if not, we really might to search further left or right of initialX.
		// For now, just throw an error
		if(barrierY == 0)
		{
			parentCarAIMain.log("[Track] findBarrier: Error, barrierY is zero, not found the start of the barrier!");
			return false;
		}
				
		// - Find X limits, by using the centre Y points, and find the extent left, then right, assign to min and max X respectively
		// Find right...
		for(x = initialX; x <= trackMaxX; x++)
		{
			// If empty, add one to barrierSizeX total, else break
			if(thresholdImage.getRGB(x, centreY) == MATCH_EMPTY_REGION) {
				barrierSizeX++;
				
				// Increase max barrierSizeY if needed
				counter = 0;
				for(y = centreY - 10; y <= centreY + 10; y++) {
					if(thresholdImage.getRGB(x, y) == MATCH_EMPTY_REGION) {
						if(y < barrierY) barrierY = y;
						counter++;
					}
				}
				if(counter > barrierSizeY) barrierSizeY = counter;
			}
			else {
				break;
			}
		}
		// Find left (and top left part
		for(x = initialX; x >= trackMinX; x--)
		{
			// If empty, add one to barrierSizeX total and set barrier X's starting point, else break
			if(thresholdImage.getRGB(x, centreY) == MATCH_EMPTY_REGION) {
				barrierSizeX++;
				barrierX = x;

				// Increase max barrierSizeY and alter barrierY if needed
				counter = 0;
				for(y = centreY - 10; y <= centreY + 10; y++) {
					if(thresholdImage.getRGB(x, y) == MATCH_EMPTY_REGION) {
						if(y < barrierY) barrierY = y;
						counter++;
					}
				}
				if(counter > barrierSizeY) barrierSizeY = counter;
			}
			else {
				break;
			}
		}

		// Report
		parentCarAIMain.logDetail("[Track] findBarrier result. barrierX [" + barrierX + "] barrierY [" + barrierY + "] barrierSizeX [" + barrierSizeX + "] barrierSizeY [" + barrierSizeY + "]");

		// Debug drawing
		//*
		camera.drawRectangle(barrierX, barrierY, barrierSizeX, barrierSizeY, debugImage);
		//*/
		return true;
	}

	// This will find if a point, x, y, is in an invalid non-track area or not
	// Faster then simply doing an RGB check - it doesn't need a thresholded image!
	public boolean pointWithinTrack(int x, int y) {
		// Have we got anything not set?
		if(trackMinX == 0) {
			parentCarAIMain.log("[Track] pointWithinTrack. Error: No valid track set.");
			return false;
		}
		if(wallMinX == 0) {
			parentCarAIMain.log("[Track] pointWithinTrack. Error: No valid walls set.");
			return false;
		}
		if(barrierX == 0) {
			parentCarAIMain.log("[Track] pointWithinTrack. Error: No valid barrier set.");
			return false;
		}
		if(chicaneX == 0) {
			parentCarAIMain.log("[Track] pointWithinTrack. Error: No valid chicane set.");
			return false;
		}
		
		// Is point out of area?
		if(x > wallMaxX || x < wallMinX || y > wallMaxY || y < wallMinY)
		{
			return false;
		}
		// Check if it is within Chicane
		if(pointIsInChicane(x, y))
		{
			return false;
		}
		// Check if it is within barrier
		if(pointIsInBarrier(x, y))
		{
			return false;
		}
		// No problems, must be on track
		return true;
	}

	// Used in pointWithinTrack only
	private boolean pointIsInChicane(int x, int y) {
		return pointIsWithinArea(x, y, chicaneX, trackMinY, chicaneSizeX, chicaneSizeY);
	}
	
	// Used in pointWithinTrack only
	private boolean pointIsInBarrier(int x, int y) {
		return pointIsWithinArea(x, y, barrierX, barrierY, barrierSizeX, barrierSizeY);
	}
	
	// Used above
	private boolean pointIsWithinArea(int x, int y, int areaX, int areaY, int areaWidth, int areaHeight) {
		// We test if it is *out* of the area...
		if(x > areaX + areaWidth || x < areaX || y > areaY + areaHeight || y < areaY)
			return false;
		return true;
	}
	
	// Set the waypoints array
	public boolean setupWaypoints(BufferedImage debugImage) {
		// Have we got anything not set?
		if(trackMinX == 0) {
			parentCarAIMain.log("[Track] setupWaypoints. Error: No valid track set.");
			return false;
		}
		if(wallMinX == 0) {
			parentCarAIMain.log("[Track] setupWaypoints. Error: No valid walls set.");
			return false;
		}
		if(barrierX == 0) {
			parentCarAIMain.log("[Track] setupWaypoints. Error: No valid barrier set.");
			return false;
		}
		if(chicaneX == 0) {
			parentCarAIMain.log("[Track] setupWaypoints. Error: No valid chicane set.");
			return false;
		}

		// Set a few waypoints around the track...
		waypointsArray = new Coordinate[8];
		// Reset (for a new race) the current waypoint number
		waypointCurrent = 0;

		// We store at the moment 8 points around the track, starting NOT from one on the starting line (that is the last point)
		int pointX, pointY, counter = -1;

		// First, we should have one 50% between chicane and the right wall, and 70% between barrier and top
		pointX = wallMaxX - ((chicaneX + chicaneSizeX)/2); // Between chicane and wall
		pointY = barrierY - Utility.getPercent(barrierY - wallMinY, 70);

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next, the right side of the track, between the barrier and the wall (70%)
		pointX = barrierX + barrierSizeX + Utility.getPercent(wallMaxX - (barrierX + barrierSizeX), 70);
		pointY = barrierY + barrierSizeY;

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next, we should have one which is in the right-hand side of the bottom
		pointX = wallMinX + Utility.getPercent(wallMaxX - wallMinX, 75);
		pointY = barrierY + barrierSizeY + Utility.getPercent(wallMaxY - (barrierY + barrierSizeY), 70);
		
		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next, we should have one which is in the middle bottom
		pointX = wallMinX + Utility.getPercent(wallMaxX - wallMinX, 50);
		pointY = barrierY + barrierSizeY + Utility.getPercent(wallMaxY - (barrierY + barrierSizeY), 70);

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next, we should have one which is in the left-hand side of the bottom
		// Note: Changed this to not 25% across, but 20% instead, so it does a better turn.
		pointX = wallMinX + Utility.getPercent(wallMaxX - wallMinX, 20);
		pointY = barrierY + barrierSizeY + Utility.getPercent(wallMaxY - (barrierY + barrierSizeY), 70);

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);
		
		// This is between the wall and the barrier on the left, 35% from the left
		pointX = barrierX - Utility.getPercent(barrierX - wallMinX, 35);
		pointY = barrierY;

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next, we should have one 50% between chicane and the left wall, and 70% up from barrier
		pointX = wallMinX + Utility.getPercent(wallMaxX - chicaneX, 50); // Between chicane and wall
		pointY = barrierY - Utility.getPercent(barrierY - wallMinY, 70);

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Next is where the car starts (or just in front) between the chicane and the barrier
		pointX = chicaneX + Utility.getPercent(chicaneSizeX, 50); // Middle of "chicane" variable
		pointY = wallMinY + chicaneSizeY + Utility.getPercent(barrierY - (wallMinY + chicaneSizeY), 50); // Half way between

		// Paint point
		waypointsArray[++counter] = new Coordinate(pointX, pointY);

		// Report outputs
		for(int index = 0; index < waypointsArray.length; index++)
		{
			parentCarAIMain.log("[Track] setupWaypoints. Waypoint [" + index + "] is at (" + waypointsArray[index].x + "," + waypointsArray[index].y + ")");
			// Debug the point on the image
			camera.drawBlock(waypointsArray[index].x, waypointsArray[index].y, debugImage, Color.PINK);
		}	
		return true;
	}
	// Get the current waypoint number
	public int getCurrentWaypoint()
	{
		return waypointCurrent;
	}
	
	// Gets the current waypoint location.
	public Coordinate getCurrentWaypointLocation()
	{
		return waypointsArray[waypointCurrent];
	}
	// Increment the waypoint
	public void incrementCurrentWaypoint()
	{
		// Array started?
		if(waypointsArray.length > 2)
		{
			// Incremenet current waypoint
			if(waypointCurrent == waypointsArray.length - 1) {
				waypointCurrent = 0;
			}
			else {
				waypointCurrent++;
			}
			// Log
			parentCarAIMain.log("[Track] Incremented current waypoint. Waypoint is now [" + waypointCurrent + "] and at (" + waypointsArray[waypointCurrent].x + "," + waypointsArray[waypointCurrent].y + ")");
		}
	}

	// Check if we need to move to the next waypoint using carPosition
	public void checkWaypoint(Coordinate carPosition)
	{	
		// Get distance between them
		double distance = Utility.getDistanceBetween(carPosition.x, carPosition.y, waypointsArray[waypointCurrent].x, waypointsArray[waypointCurrent].y);
		
		parentCarAIMain.logDetail("[Track] checkWaypoint. Distance between (" + carPosition.x + "," + carPosition.y + ") and (" + waypointsArray[waypointCurrent].x + "," + waypointsArray[waypointCurrent].y + ") is " + distance);

		// Distance needed to register
		// Note: the direction of the car isn't taken into account, sadly.
		double distanceNeeded = 16.0;
		// Within a small distance?
		if(distance <= distanceNeeded)
		{
			// Incremenet current waypoint
			if(waypointCurrent == waypointsArray.length - 1) {
				waypointCurrent = 0;
			}
			else {
				waypointCurrent++;
			}
			parentCarAIMain.log("[Track] checkWaypoint. Moved waypoint on. Waypoint is now [" + waypointCurrent + "] and at (" + waypointsArray[waypointCurrent].x + "," + waypointsArray[waypointCurrent].y + ")");
		}
	}

	/*************************
		PRIVATE UTILITY FUNCTIONS 
	*************************/

	// Checks the a horizontal line "y" for a color, "colorMatch" between minX and maxX
	// Can optionally have a gap and percent needed of this colour
	private boolean scanHorizontalLineForColor(int y, int minX, int maxX, int colorMatch, BufferedImage image) { return scanHorizontalLineForColor(y, minX, maxX, colorMatch, image, 100, -1, -1);  }
	private boolean scanHorizontalLineForColor(int y, int minX, int maxX, int colorMatch, BufferedImage image, int percent) { return scanHorizontalLineForColor(y, minX, maxX, colorMatch, image, percent, -1, -1);  }
	// Full implementation
	private boolean scanHorizontalLineForColor(int y, int minX, int maxX, int colorMatch, BufferedImage image, int percent, int ignoreMinX, int ignoreMaxX)
	{
		// For the % check
		double totalOfColour = 0.0, totalPixels = 0.0;

		// Error checking
		if(y >= image.getHeight()) y = image.getHeight() - 1;
		if(y < 0) y = 0;
		
		// Look between these co-ordinates
		for(int x = minX; x < maxX; x++)
		{
			// Cannot be within the boundried point
			if(x < ignoreMinX || x > ignoreMaxX)
			{
				// Check color - add up pixels
				totalPixels++;
				if(image.getRGB(x,y) == colorMatch)
				{
					totalOfColour++;
				}
			}
		}
		if(totalOfColour > 0)
		{
			// Do we have the % work? (integer division)
			double percentLine = (totalOfColour / totalPixels) * 100;
			if(percentLine >= percent)
			{
				parentCarAIMain.logDetail("[Track] scanHorizontalLineForColor: found scanline at y: " + y + ". Percent: " + percentLine + " out of " + percent + ". (totalOfColour = " + totalOfColour + ". totalPixels = " + totalPixels);
				return true;
			}
			else
			{
				parentCarAIMain.logDetail("[Track] scanHorizontalLineForColor: not found scanline at y: " + y + ". Percent: " + percentLine + " out of " + percent + ". (totalOfColour = " + totalOfColour + ". totalPixels = " + totalPixels);
				return false;
			}
		}
		parentCarAIMain.logDetail("[Track] scanHorizontalLineForColor: not found scanline at y: " + y + ". Percent: none! out of " + percent + ".");
		// Not found on this line
		return false;
	}
	
	// Same as above one, but instead is vertical on line X
	private boolean scanVerticalLineForColor(int x, int minY, int maxY, int colorMatch, BufferedImage image) { return scanVerticalLineForColor(x, minY, maxY, colorMatch, image, 100, -1, -1);  }
	private boolean scanVerticalLineForColor(int x, int minY, int maxY, int colorMatch, BufferedImage image, int percent) { return scanVerticalLineForColor(x, minY, maxY, colorMatch, image, percent, -1, -1);  }
	private boolean scanVerticalLineForColor(int x, int minY, int maxY, int colorMatch, BufferedImage image, int percent, int ignoreMinY, int ignoreMaxY)
	{
		double totalPixels = 0.0, totalOfColour = 0.0;
	
		// Error checking
		if(x >= image.getWidth()) x = image.getWidth() - 1;
		if(x < 0) x = 0;
	
		// Look between these co-ordinates
		for(int y = minY; y < maxY; y++)
		{
			// Cannot be within the boundried point
			if(y < ignoreMinY || y > ignoreMaxY)
			{
				// Check color - add up pixels
				totalPixels++;
				if(image.getRGB(x,y) == colorMatch)
				{
					totalOfColour++;
				}
			}
		}
		if(totalOfColour > 0)
		{		
			// Do we have the % work? (integer division)
			double percentLine = (totalOfColour / totalPixels) * 100;
			if(percentLine  >= percent)
			{
				parentCarAIMain.logDetail("[Track] scanVerticalLineForColor: found scanline at x: " + x + ". Percent: " + percentLine + " out of " + percent + ". (totalOfColour = " + totalOfColour + ". totalPixels = " + totalPixels);
				return true;
			}
			else
			{
				parentCarAIMain.logDetail("[Track] scanVerticalLineForColor: not found scanline at x: " + x + ". Percent: " + percentLine + " out of " + percent + ". (totalOfColour = " + totalOfColour + ". totalPixels = " + totalPixels);
				return false;
			}
		}
		parentCarAIMain.logDetail("[Track] scanVerticalLineForColor: not found scanline at x: " + x + ". Percent: none! out of " + percent + ".");
		// Not found on this line
		return false;
	}
}